# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tensorflow import keras

from keras_tuner import utils
from keras_tuner.engine.hyperparameters import hp_types
from keras_tuner.engine.hyperparameters.hp_types import Boolean
from keras_tuner.engine.hyperparameters.hp_types import Choice
from keras_tuner.engine.hyperparameters.hp_types import Fixed
from keras_tuner.engine.hyperparameters.hp_types import Float
from keras_tuner.engine.hyperparameters.hp_types import Int
from keras_tuner.engine.hyperparameters.hyperparameter import HyperParameter
from keras_tuner.engine.hyperparameters.hyperparameters import HyperParameters

OBJECTS = hp_types.OBJECTS + (
    HyperParameter,
    HyperParameters,
)

ALL_CLASSES = {cls.__name__: cls for cls in OBJECTS}


def deserialize(config):
    return utils.deserialize_keras_object(config, module_objects=ALL_CLASSES)


def serialize(obj):
    return utils.serialize_keras_object(obj)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Distribution utilities."""

import os


def has_chief_oracle():
    """Checks for distributed tuning with a chief Oracle.

    `CloudOracle` manages its own distribution so should not set
    "KERASTUNER_ORACLE_IP".

    Returns:
        Boolean, whether distributed tuning with a chief Oracle should be run.
    """
    if "KERASTUNER_ORACLE_IP" in os.environ:
        if "KERASTUNER_ORACLE_PORT" not in os.environ:
            raise RuntimeError(
                'Environment variable "KERASTUNER_ORACLE_IP" was set, '
                'but "KERASTUNER_ORACLE_PORT" was not. Please specify '
                "a port."
            )
        if "KERASTUNER_TUNER_ID" not in os.environ:
            raise RuntimeError(
                'Environment variable "KERASTUNER_ORACLE_IP" was set, '
                'but "KERASTUNER_TUNER_ID" was not. Please specify '
                "an ID for each tuner."
            )
        return True
    return False


def is_chief_oracle():
    if has_chief_oracle():
        return "chief" in os.environ["KERASTUNER_TUNER_ID"]
    return False



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""KerasTuner utilities."""


import warnings

import tensorflow as tf
from packaging.version import parse

# Check if we are in a ipython/colab environement
try:
    import IPython

    class_name = IPython.get_ipython().__class__.__name__
    IS_NOTEBOOK = "Terminal" not in class_name
except (NameError, ImportError):  # pragma: no cover
    IS_NOTEBOOK = False  # pragma: no cover


if IS_NOTEBOOK:
    from IPython import display


def try_clear():
    if IS_NOTEBOOK:
        display.clear_output()
    else:
        print()


def create_directory(path, remove_existing=False):
    # Create the directory if it doesn't exist.
    if not tf.io.gfile.exists(path):
        tf.io.gfile.makedirs(path)

    # If it does exist, and remove_existing is specified,
    # the directory will be removed and recreated.
    elif remove_existing:
        tf.io.gfile.rmtree(path)
        tf.io.gfile.makedirs(path)


def check_tf_version():
    if parse(tf.__version__) < parse("2.0.0"):
        warnings.warn(
            "The Tensorflow package version needs to be at least 2.0.0 \n"
            "for AutoKeras to run. Currently, your TensorFlow version is \n"
            f"{tf.__version__}. Please upgrade with \n"
            "`$ pip install --upgrade tensorflow`. \n"
            "You can use `pip freeze` to check afterwards that everything is "
            "ok.",
            ImportWarning,
        )


def serialize_keras_object(obj):
    if hasattr(tf.keras.utils, "legacy"):
        return tf.keras.utils.legacy.serialize_keras_object(  # pragma: no cover
            obj
        )
    else:
        return tf.keras.utils.serialize_keras_object(obj)  # pragma: no cover


def deserialize_keras_object(config, module_objects=None, custom_objects=None):
    if hasattr(tf.keras.utils, "legacy"):
        return (  # pragma: no cover
            tf.keras.utils.legacy.deserialize_keras_object(
                config, custom_objects, module_objects
            )
        )
    else:
        return tf.keras.utils.deserialize_keras_object(  # pragma: no cover
            config, custom_objects, module_objects
        )


def to_list(values):
    if isinstance(values, list):
        return values
    if isinstance(values, tuple):
        return list(values)
    return [values]



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tuner for Scikit-learn Models."""
import collections
import inspect
import os
import pickle

import numpy as np
import tensorflow as tf

try:
    import pandas as pd  # pytype: disable=import-error
except ImportError:  # pragma: no cover
    pd = None  # pragma: no cover

try:
    import sklearn  # pytype: disable=import-error
    import sklearn.model_selection
    import sklearn.pipeline
except ImportError:  # pragma: no cover
    sklearn = None  # pragma: no cover

from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import base_tuner


def split_data(data, indices):
    if isinstance(data, np.ndarray):
        return data[indices]
    elif pd is not None and isinstance(data, pd.DataFrame):
        return data.iloc[indices]
    else:
        raise TypeError(
            "Expected the data to be numpy.ndarray or pandas.DataFrame. "
            f"Received: {data}."
        )


@keras_tuner_export(
    ["keras_tuner.SklearnTuner", "keras_tuner.tuners.SklearnTuner"]
)
class SklearnTuner(base_tuner.BaseTuner):
    """Tuner for Scikit-learn Models.

    Performs cross-validated hyperparameter search for Scikit-learn models.

    Examples:

    ```python
    import keras_tuner
    from sklearn import ensemble
    from sklearn import datasets
    from sklearn import linear_model
    from sklearn import metrics
    from sklearn import model_selection

    def build_model(hp):
      model_type = hp.Choice('model_type', ['random_forest', 'ridge'])
      if model_type == 'random_forest':
        model = ensemble.RandomForestClassifier(
            n_estimators=hp.Int('n_estimators', 10, 50, step=10),
            max_depth=hp.Int('max_depth', 3, 10))
      else:
        model = linear_model.RidgeClassifier(
            alpha=hp.Float('alpha', 1e-3, 1, sampling='log'))
      return model

    tuner = keras_tuner.tuners.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective('score', 'max'),
            max_trials=10),
        hypermodel=build_model,
        scoring=metrics.make_scorer(metrics.accuracy_score),
        cv=model_selection.StratifiedKFold(5),
        directory='.',
        project_name='my_project')

    X, y = datasets.load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = model_selection.train_test_split(
        X, y, test_size=0.2)

    tuner.search(X_train, y_train)

    best_model = tuner.get_best_models(num_models=1)[0]
    ```

    Args:
        oracle: A `keras_tuner.Oracle` instance. Note that for this `Tuner`,
            the `objective` for the `Oracle` should always be set to
            `Objective('score', direction='max')`. Also, `Oracle`s that exploit
            Neural-Network-specific training (e.g. `Hyperband`) should not be
            used with this `Tuner`.
        hypermodel: A `HyperModel` instance (or callable that takes
            hyperparameters and returns a Model instance).
        scoring: An sklearn `scoring` function. For more information, see
            `sklearn.metrics.make_scorer`. If not provided, the Model's default
            scoring will be used via `model.score`. Note that if you are
            searching across different Model families, the default scoring for
            these Models will often be different. In this case you should
            supply `scoring` here in order to make sure your Models are being
            scored on the same metric.
        metrics: Additional `sklearn.metrics` functions to monitor during
            search. Note that these metrics do not affect the search process.
        cv: An `sklearn.model_selection` Splitter class. Used to
            determine how samples are split up into groups for
            cross-validation.
        **kwargs: Keyword arguments relevant to all `Tuner` subclasses. Please
            see the docstring for `Tuner`.
    """

    def __init__(
        self, oracle, hypermodel, scoring=None, metrics=None, cv=None, **kwargs
    ):
        super().__init__(oracle=oracle, hypermodel=hypermodel, **kwargs)

        if sklearn is None:
            raise ImportError(
                "Please install sklearn before using the `SklearnTuner`."
            )

        self.scoring = scoring

        if metrics is None:
            metrics = []
        if not isinstance(metrics, (list, tuple)):
            metrics = [metrics]
        self.metrics = metrics

        self.cv = cv or sklearn.model_selection.KFold(
            5, shuffle=True, random_state=1
        )

    def search(self, X, y, sample_weight=None, groups=None):
        """Performs hyperparameter search.

        Args:
            X: See docstring for `model.fit` for the `sklearn` Models being
                tuned.
            y: See docstring for `model.fit` for the `sklearn` Models being
                tuned.
            sample_weight: Optional. See docstring for `model.fit` for the
                `sklearn` Models being tuned.
            groups: Optional. Required for `sklearn.model_selection` Splitter
                classes that split based on group labels (For example, see
                `sklearn.model_selection.GroupKFold`).
        """
        # Only overridden for the docstring.
        return super().search(X, y, sample_weight=sample_weight, groups=groups)

    def run_trial(self, trial, X, y, sample_weight=None, groups=None):
        metrics = collections.defaultdict(list)
        # For cross-validation methods that expect a `groups` argument.
        cv_kwargs = {"groups": groups} if groups is not None else {}
        for train_indices, test_indices in self.cv.split(X, y, **cv_kwargs):
            X_train = split_data(X, train_indices)
            y_train = split_data(y, train_indices)
            X_test = split_data(X, test_indices)
            y_test = split_data(y, test_indices)

            sample_weight_train = (
                sample_weight[train_indices]
                if sample_weight is not None
                else None
            )

            model = self.hypermodel.build(trial.hyperparameters)

            supports_sw = (
                "sample_weight" in inspect.getfullargspec(model.fit).args
            )
            if isinstance(model, sklearn.pipeline.Pipeline) or not supports_sw:
                model.fit(X_train, y_train)
            else:
                model.fit(X_train, y_train, sample_weight=sample_weight_train)

            sample_weight_test = (
                sample_weight[test_indices]
                if sample_weight is not None
                else None
            )

            if self.scoring is None:
                score = model.score(
                    X_test, y_test, sample_weight=sample_weight_test
                )
            else:
                score = self.scoring(
                    model, X_test, y_test, sample_weight=sample_weight_test
                )
            metrics["score"].append(score)

            if self.metrics:
                y_test_pred = model.predict(X_test)
                for metric in self.metrics:
                    result = metric(
                        y_test, y_test_pred, sample_weight=sample_weight_test
                    )
                    metrics[metric.__name__].append(result)

        self.save_model(trial.trial_id, model)
        return {name: np.mean(values) for name, values in metrics.items()}

    def save_model(self, trial_id, model, step=0):
        fname = os.path.join(self.get_trial_dir(trial_id), "model.pickle")
        with tf.io.gfile.GFile(fname, "wb") as f:
            pickle.dump(model, f)

    def load_model(self, trial):
        fname = os.path.join(self.get_trial_dir(trial.trial_id), "model.pickle")
        with tf.io.gfile.GFile(fname, "rb") as f:
            return pickle.load(f)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

from keras_tuner import utils
from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import conditions as conditions_mod


@keras_tuner_export("keras_tuner.engine.hyperparameters.HyperParameter")
class HyperParameter:
    """Hyperparameter base class.

    A `HyperParameter` instance is uniquely identified by its `name` and
    `conditions` attributes. `HyperParameter`s with the same `name` but with
    different `conditions` are considered as different `HyperParameter`s by
    the `HyperParameters` instance.

    Args:
        name: A string. the name of parameter. Must be unique for each
            `HyperParameter` instance in the search space.
        default: The default value to return for the parameter.
        conditions: A list of `Condition`s for this object to be considered
            active.
    """

    def __init__(self, name, default=None, conditions=None):
        self.name = name
        self._default = default

        conditions = utils.to_list(conditions) if conditions else []
        self.conditions = conditions

    def get_config(self):
        conditions = [conditions_mod.serialize(c) for c in self.conditions]
        return {
            "name": self.name,
            "default": self.default,
            "conditions": conditions,
        }

    @property
    def default(self):
        return self._default

    @property
    def values(self):
        """Return a iterable of all possible values of the hp."""
        raise NotImplementedError

    def random_sample(self, seed=None):
        random_state = random.Random(seed)
        prob = float(random_state.random())
        return self.prob_to_value(prob)

    def prob_to_value(self, prob):
        """Convert cumulative probability in range [0.0, 1.0) to hp value."""
        raise NotImplementedError

    def value_to_prob(self, value):
        """Convert a hp value to cumulative probability in range [0.0, 1.0)."""
        raise NotImplementedError

    @classmethod
    def from_config(cls, config):
        config["conditions"] = [
            conditions_mod.deserialize(c) for c in config["conditions"]
        ]
        return cls(**config)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import numpy as np
import pytest
import tensorflow as tf


@pytest.fixture(autouse=True)
def set_seeds_before_tests():
    """Test wrapper to set the seed before each test.

    This wrapper runs for all the tests in the test suite.
    """
    random.seed(0)
    np.random.seed(0)
    tf.random.set_seed(0)
    yield



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import keras_tuner
from keras_tuner.engine import objective


def test_create_objective_with_str():
    obj = objective.create_objective("accuracy")
    assert obj.name == "accuracy" and obj.direction == "max"


def test_create_objective_with_objective():
    obj = objective.create_objective("accuracy")
    obj = objective.create_objective(keras_tuner.Objective("score", "min"))
    assert obj.name == "score" and obj.direction == "min"


def test_create_objective_with_multi_objective():
    obj = objective.create_objective(
        [
            keras_tuner.Objective("score", "max"),
            keras_tuner.Objective("loss", "min"),
        ]
    )
    assert isinstance(obj, objective.MultiObjective)
    assert (
        obj.objectives[0].name == "score"
        and obj.objectives[0].direction == "max"
    )
    assert (
        obj.objectives[1].name == "loss"
        and obj.objectives[1].direction == "min"
    )


def test_create_objective_with_multi_str():
    obj = objective.create_objective(["accuracy", "loss"])
    assert isinstance(obj, objective.MultiObjective)
    assert (
        obj.objectives[0].name == "accuracy"
        and obj.objectives[0].direction == "max"
    )
    assert (
        obj.objectives[1].name == "loss"
        and obj.objectives[1].direction == "min"
    )


def test_objective_better_than_max():
    obj = objective.create_objective("accuracy")
    assert obj.better_than(1, 0)
    assert not obj.better_than(0, 1)
    assert not obj.better_than(0, 0)


def test_objective_better_than_min():
    obj = objective.create_objective("loss")
    assert obj.better_than(0, 1)
    assert not obj.better_than(1, 0)
    assert not obj.better_than(0, 0)


def test_objective_has_value():
    obj = objective.create_objective("loss")
    assert obj.has_value({"loss": 3.0})
    assert not obj.has_value({"accuracy": 3.0})


def test_objective_get_value():
    obj = objective.create_objective("loss")
    assert obj.get_value({"accuracy": 3.0, "loss": 2.0}) == 2.0


def test_multi_objective_get_value():
    obj = objective.create_objective(["accuracy", "loss"])
    assert obj.get_value({"accuracy": 3.0, "loss": 2.0}) == -1.0


def test_objective_equal():
    obj1 = objective.Objective(name="accuracy", direction="max")
    obj2 = objective.Objective(name="accuracy", direction="max")
    assert obj1 == obj2


def test_objective_not_equal_with_diff_name():
    obj1 = objective.Objective(name="accuracy1", direction="max")
    obj2 = objective.Objective(name="accuracy", direction="max")
    assert obj1 != obj2


def test_objective_not_equal_with_diff_dir():
    obj1 = objective.Objective(name="accuracy", direction="min")
    obj2 = objective.Objective(name="accuracy", direction="max")
    assert obj1 != obj2


def test_multi_objective_equal():
    obj1 = objective.create_objective(["accuracy", "loss"])
    obj2 = objective.create_objective(["loss", "accuracy"])
    assert obj1 == obj2


def test_multi_objective_not_equal():
    obj1 = objective.create_objective(["loss", "loss"])
    obj2 = objective.create_objective(["loss", "accuracy"])
    assert obj1 != obj2


def test_multi_objective_has_value():
    obj = objective.create_objective(["loss", "accuracy"])
    assert obj.has_value({"loss": 1.0, "accuracy": 1.0, "mse": 2.0})
    assert not obj.has_value({"accuracy": 1.0, "mse": 2.0})


def test_objective_to_str():
    obj = objective.Objective(name="accuracy", direction="min")
    assert str(obj) == 'Objective(name="accuracy", direction="min")'


def test_multi_objective_to_str():
    obj = objective.create_objective(["loss", "accuracy"])
    assert (
        str(obj) == 'MultiObjective(name="multi_objective", direction="min"): '
        '[Objective(name="loss", direction="min"), '
        'Objective(name="accuracy", direction="max")]'
    )


def test_unknown_objective_type_error():
    with pytest.raises(TypeError, match="not understood, expected str"):
        objective.create_objective([3, "accuracy"])



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import numpy as np
import pytest
from tensorflow.keras import losses
from tensorflow.keras import metrics

from keras_tuner.engine import metrics_tracking


def test_register_from_metrics():
    # As well as direction inference.
    tracker = metrics_tracking.MetricsTracker(
        metrics=[metrics.CategoricalAccuracy(), metrics.MeanSquaredError()]
    )
    assert set(tracker.metrics.keys()) == {
        "categorical_accuracy",
        "mean_squared_error",
    }
    assert tracker.metrics["categorical_accuracy"].direction == "max"
    assert tracker.metrics["mean_squared_error"].direction == "min"


def test_register():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("new_metric", direction="max")
    assert set(tracker.metrics.keys()) == {"new_metric"}
    assert tracker.metrics["new_metric"].direction == "max"
    with pytest.raises(ValueError, match="`direction` should be one of"):
        tracker.register("another_metric", direction="wrong")
    with pytest.raises(ValueError, match="already exists"):
        tracker.register("new_metric", direction="max")


def test_exists():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("new_metric", direction="max")
    assert tracker.exists("new_metric")
    assert not tracker.exists("another_metric")


def test_update():
    tracker = metrics_tracking.MetricsTracker()
    tracker.update("new_metric", 0.5)  # automatic registration
    assert set(tracker.metrics.keys()) == {"new_metric"}
    assert tracker.metrics["new_metric"].direction == "min"  # default direction
    assert tracker.get_history("new_metric") == [
        metrics_tracking.MetricObservation(0.5, step=0)
    ]


def test_metric_observation_repr():
    assert (
        repr(metrics_tracking.MetricObservation(0.5, step=0))
        == "MetricObservation(value=[0.5], step=0)"
    )


def test_get_history():
    tracker = metrics_tracking.MetricsTracker()
    tracker.update("new_metric", 0.5, step=0)
    tracker.update("new_metric", 1.5, step=1)
    tracker.update("new_metric", 2.0, step=2)
    assert tracker.get_history("new_metric") == [
        metrics_tracking.MetricObservation(0.5, 0),
        metrics_tracking.MetricObservation(1.5, 1),
        metrics_tracking.MetricObservation(2.0, 2),
    ]
    with pytest.raises(ValueError, match="Unknown metric"):
        tracker.get_history("another_metric")


def test_set_history():
    tracker = metrics_tracking.MetricsTracker()
    tracker.set_history(
        "new_metric",
        [
            metrics_tracking.MetricObservation(0.5, 0),
            metrics_tracking.MetricObservation(1.5, 1),
            metrics_tracking.MetricObservation(2.0, 2),
        ],
    )
    values = [obs.value for obs in tracker.get_history("new_metric")]
    steps = [obs.step for obs in tracker.get_history("new_metric")]
    assert values == [[0.5], [1.5], [2.0]]
    assert steps == [0, 1, 2]


def test_get_best_step_value_none():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("val_loss", "min")
    assert tracker.get_best_step("val_loss") is None


def test_get_best_value():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("metric_min", "min")
    tracker.register("metric_max", "max")
    assert tracker.get_best_value("metric_min") is None

    tracker.set_history(
        "metric_min",
        [
            metrics_tracking.MetricObservation(1.0, 0),
            metrics_tracking.MetricObservation(2.0, 1),
            metrics_tracking.MetricObservation(3.0, 2),
        ],
    )
    tracker.set_history(
        "metric_max",
        [
            metrics_tracking.MetricObservation(1.0, 0),
            metrics_tracking.MetricObservation(2.0, 1),
            metrics_tracking.MetricObservation(3.0, 2),
        ],
    )
    assert tracker.get_best_value("metric_min") == 1.0
    assert tracker.get_best_value("metric_max") == 3.0


def test_get_statistics():
    tracker = metrics_tracking.MetricsTracker()
    history = [
        metrics_tracking.MetricObservation(random.random(), i)
        for i in range(10)
    ]
    tracker.set_history("new_metric", history)
    stats = tracker.get_statistics("new_metric")
    assert set(stats.keys()) == {"min", "max", "mean", "median", "var", "std"}
    history = [obs.value for obs in history]
    assert stats["min"] == np.min(history)
    assert stats["max"] == np.max(history)
    assert stats["mean"] == np.mean(history)
    assert stats["median"] == np.median(history)
    assert stats["var"] == np.var(history)
    assert stats["std"] == np.std(history)


def test_get_last_value():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("new_metric", "min")
    assert tracker.get_last_value("new_metric") is None
    tracker.set_history(
        "new_metric",
        [
            metrics_tracking.MetricObservation(1.0, 0),
            metrics_tracking.MetricObservation(2.0, 1),
            metrics_tracking.MetricObservation(3.0, 2),
        ],
    )
    assert tracker.get_last_value("new_metric") == 3.0


def test_serialization():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("metric_min", "min")
    tracker.register("metric_max", "max")
    tracker.set_history(
        "metric_min",
        [
            metrics_tracking.MetricObservation(1.0, 0),
            metrics_tracking.MetricObservation(2.0, 1),
            metrics_tracking.MetricObservation(3.0, 2),
        ],
    )
    tracker.set_history(
        "metric_max",
        [
            metrics_tracking.MetricObservation(1.0, 0),
            metrics_tracking.MetricObservation(2.0, 1),
            metrics_tracking.MetricObservation(3.0, 2),
        ],
    )

    new_tracker = metrics_tracking.MetricsTracker.from_config(
        tracker.get_config()
    )
    assert new_tracker.metrics.keys() == tracker.metrics.keys()


def test_metricobservation_proto():
    obs = metrics_tracking.MetricObservation(-10, 5)
    proto = obs.to_proto()
    assert proto.value == [-10]
    assert proto.step == 5
    new_obs = metrics_tracking.MetricObservation.from_proto(proto)
    assert new_obs == obs


def test_metrichistory_proto():
    tracker = metrics_tracking.MetricHistory("max")
    tracker.update(5, step=3)
    tracker.update(10, step=4)

    proto = tracker.to_proto()
    assert proto.maximize
    assert proto.observations[0].value == [5]
    assert proto.observations[0].step == 3
    assert proto.observations[1].value == [10]
    assert proto.observations[1].step == 4

    new_tracker = metrics_tracking.MetricHistory.from_proto(proto)
    assert new_tracker.direction == "max"
    assert new_tracker.get_history() == [
        metrics_tracking.MetricObservation(5, 3),
        metrics_tracking.MetricObservation(10, 4),
    ]


def test_metricstracker_proto():
    tracker = metrics_tracking.MetricsTracker()
    tracker.register("score", direction="max")
    tracker.update("score", value=10, step=1)
    tracker.update("score", value=20, step=1)
    tracker.update("score", value=30, step=2)

    proto = tracker.to_proto()
    obs = proto.metrics["score"].observations
    assert obs[0].value == [10, 20]
    assert obs[0].step == 1
    assert obs[1].value == [30]
    assert obs[1].step == 2
    assert proto.metrics["score"].maximize

    new_tracker = metrics_tracking.MetricsTracker.from_proto(proto)
    assert new_tracker.metrics["score"].direction == "max"
    assert new_tracker.metrics["score"].get_history() == [
        metrics_tracking.MetricObservation([10, 20], 1),
        metrics_tracking.MetricObservation(30, 2),
    ]


def test_metric_direction_inference():
    # Test min metrics.
    assert metrics_tracking.infer_metric_direction("MAE") == "min"
    assert (
        metrics_tracking.infer_metric_direction(metrics.binary_crossentropy)
        == "min"
    )
    assert (
        metrics_tracking.infer_metric_direction(metrics.FalsePositives())
        == "min"
    )

    # All losses in keras.losses are considered as 'min'.
    assert metrics_tracking.infer_metric_direction("squared_hinge") == "min"
    assert metrics_tracking.infer_metric_direction(losses.hinge) == "min"
    assert (
        metrics_tracking.infer_metric_direction(
            losses.CategoricalCrossentropy()
        )
        == "min"
    )

    # Test max metrics.
    assert metrics_tracking.infer_metric_direction("binary_accuracy") == "max"
    assert (
        metrics_tracking.infer_metric_direction(metrics.categorical_accuracy)
        == "max"
    )
    assert metrics_tracking.infer_metric_direction(metrics.Precision()) == "max"

    # Test unknown metrics.
    assert metrics_tracking.infer_metric_direction("my_metric") is None

    def my_metric_fn(x, y):
        return x

    assert metrics_tracking.infer_metric_direction(my_metric_fn) is None

    class MyMetric(metrics.Metric):
        def update_state(self, x, y):
            return 1

        def result(self):
            return 1

    assert metrics_tracking.infer_metric_direction(MyMetric()) is None

    # Test special cases.
    assert metrics_tracking.infer_metric_direction("loss") == "min"
    assert metrics_tracking.infer_metric_direction("acc") == "max"
    assert metrics_tracking.infer_metric_direction("val_acc") == "max"
    assert metrics_tracking.infer_metric_direction("crossentropy") == "min"
    assert metrics_tracking.infer_metric_direction("ce") == "min"
    assert metrics_tracking.infer_metric_direction("weighted_acc") == "max"
    assert metrics_tracking.infer_metric_direction("val_weighted_ce") == "min"
    assert (
        metrics_tracking.infer_metric_direction("weighted_binary_accuracy")
        == "max"
    )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

from tensorflow import keras

from keras_tuner.engine import hypermodel
from keras_tuner.engine import trial as trial_module
from keras_tuner.tuners import gridsearch


def test_that_exhaustive_space_is_explored(tmp_path):
    # Tests that it explores the whole search space given by the combination
    # of all hyperparameter of Choice type.

    # Given
    want_units_1 = [2, 4]
    want_units_2 = [1, 3]
    want_optimizers = ["adam", "sgd", "rmsprop"]
    want_loss = "binary_crossentropy"
    want_dropouts = [True, False]

    class MyGridSearch(gridsearch.GridSearch):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            hp.Choice("units_1", values=want_units_1)
            hp.Boolean("dropout", default=want_dropouts[0])
            hp.Choice("units_2", values=want_units_2)
            hp.Choice("optmizer", values=want_optimizers),
            hp.Fixed("loss", value=want_loss)
            return random.random()

    # When
    tuner = MyGridSearch(directory=tmp_path)
    tuner.search()

    # Then
    assert {hp.name for hp in tuner.oracle.get_space().space} == {
        "units_1",
        "optmizer",
        "units_2",
        "loss",
        "dropout",
    }

    # 2 units_1, 3 optimizers, 2 units_2, 2 dropout and 1 loss
    expected_hyperparameter_space = 24
    assert len(tuner.oracle.trials) == expected_hyperparameter_space

    trials = tuner.oracle.get_best_trials(
        num_trials=expected_hyperparameter_space
    )
    explored_space = [trial.hyperparameters.values for trial in trials]
    for want_unit_1 in want_units_1:
        for want_unit_2 in want_units_2:
            for want_optimizer in want_optimizers:
                for want_dropout in want_dropouts:
                    assert {
                        "units_1": want_unit_1,
                        "units_2": want_unit_2,
                        "optmizer": want_optimizer,
                        "loss": want_loss,
                        "dropout": want_dropout,
                    } in explored_space


def test_int_and_float(tmp_path):
    class MyGridSearch(gridsearch.GridSearch):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            hp.Int("int", 1, 5)
            hp.Float("float", 1, 2)
            return random.random()

    tuner = MyGridSearch(directory=tmp_path)
    tuner.search()
    # int has 5 values, float sampled 10 values and 1 default value
    # 5 * (10 + 1)
    assert len(tuner.oracle.trials) == 55


def test_new_hp(tmp_path):
    class MyGridSearch(gridsearch.GridSearch):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            if hp.Boolean("bool"):
                hp.Choice("choice1", [0, 1, 2])
            else:
                hp.Choice("choice2", [3, 4, 5])
            return random.random()

    tuner = MyGridSearch(directory=tmp_path)
    tuner.search(verbose=0)
    assert len(tuner.oracle.trials) == 3 + 3 * 3


def test_hp_in_fit(tmp_path):
    class MyHyperModel(hypermodel.HyperModel):
        def build(self, hp):
            hp.Fixed("fixed", 3)
            return keras.Sequential()

        def fit(self, hp, model, *args, **kwargs):
            hp.Choice("choice", [0, 1, 2])
            return random.random()

    tuner = gridsearch.GridSearch(hypermodel=MyHyperModel(), directory=tmp_path)
    tuner.search(verbose=0)
    assert len(tuner.oracle.trials) == 3


def test_conditional_scope(tmp_path):
    class MyHyperModel(hypermodel.HyperModel):
        def build(self, hp):
            a = hp.Boolean("bool")
            with hp.conditional_scope("bool", [True]):
                if a:
                    hp.Choice("choice1", [1, 2])
            with hp.conditional_scope("bool", [False]):
                if not a:
                    hp.Choice("choice2", [3, 4])
            return keras.Sequential()

        def fit(self, hp, model, *args, **kwargs):
            a = hp.Boolean("bool2")
            with hp.conditional_scope("bool2", [True]):
                if a:
                    hp.Choice("choice3", [1, 2])
            with hp.conditional_scope("bool2", [False]):
                if not a:
                    hp.Choice("choice4", [3, 4])
            return random.random()

    tuner = gridsearch.GridSearch(hypermodel=MyHyperModel(), directory=tmp_path)
    tuner.search(verbose=0)
    assert len(tuner.oracle.trials) == 4 * 4


def test_exhaust_trials_in_between_before_the_latter_finishes(tmp_path):
    class MyHyperModel(hypermodel.HyperModel):
        def build(self, hp):
            hp.Boolean("bool")
            return keras.Sequential()

        def fit(self, hp, model, *args, **kwargs):
            hp.Boolean("bool2")
            return random.random()

    tuner = gridsearch.GridSearch(hypermodel=MyHyperModel(), directory=tmp_path)
    oracle = tuner.oracle

    def run(trial):
        hp = trial.hyperparameters
        hm = MyHyperModel()
        hm.fit(hp, hm.build(hp))
        oracle.update_space(hp)

    def end_trial(trial):
        run(trial)
        oracle.update_trial(
            trial_id=trial.trial_id,
            metrics={oracle.objective.name: random.random()},
        )
        trial.status = trial_module.TrialStatus.COMPLETED
        oracle.end_trial(trial)

    trial_1 = oracle.create_trial(tuner_id="1")
    assert trial_1.status == trial_module.TrialStatus.RUNNING
    trial_2 = oracle.create_trial(tuner_id="2")
    assert trial_2.status == trial_module.TrialStatus.RUNNING

    # Iterated bool1, bool2 not discovered yet. So idle.
    trial_3 = oracle.create_trial(tuner_id="3")
    assert trial_3.status == trial_module.TrialStatus.IDLE

    end_trial(trial_1)
    # Discovered bool2 in trial_1, so new value of bool2 for trial_3 after
    # trial_1.
    trial_3 = oracle.create_trial(tuner_id="3")
    assert trial_3.status == trial_module.TrialStatus.RUNNING
    # Exhausted all possible combinations whose order is between trial_1 and
    # trial_2.  So idle.
    trial_4 = oracle.create_trial(tuner_id="4")
    assert trial_4.status == trial_module.TrialStatus.IDLE

    end_trial(trial_2)
    # New value of bool2 for trial_4 after trial_2.
    trial_4 = oracle.create_trial(tuner_id="4")
    assert trial_4.status == trial_module.TrialStatus.RUNNING
    trial_5 = oracle.create_trial(tuner_id="5")
    assert trial_5.status == trial_module.TrialStatus.IDLE

    end_trial(trial_3)
    end_trial(trial_4)
    trial_5 = oracle.create_trial(tuner_id="5")
    assert trial_5.status == trial_module.TrialStatus.STOPPED


def test_linked_list():
    linked_list = gridsearch.LinkedList()
    linked_list.insert("0")
    assert linked_list.next("0") is None
    linked_list.insert("1")
    assert linked_list.next("0") == "1"
    assert linked_list.next("1") is None
    linked_list.insert("2", "0")
    assert linked_list.next("0") == "2"
    assert linked_list.next("2") == "1"
    assert linked_list.next("1") is None
    linked_list.insert("3", "1")
    linked_list.insert("4")
    assert linked_list.next("0") == "2"
    assert linked_list.next("2") == "1"
    assert linked_list.next("1") == "3"
    assert linked_list.next("3") == "4"
    assert linked_list.next("4") is None



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

import numpy as np
import pytest

import keras_tuner
from keras_tuner.engine import oracle as oracle_module
from keras_tuner.engine import trial as trial_module


class OracleStub(oracle_module.Oracle):
    def __init__(self, directory, **kwargs):
        super().__init__(**kwargs)
        self.score_trial_called = False
        self._set_project_dir(directory=directory, project_name="name")

    def populate_space(self, trial_id):
        return {
            "values": {"hp1": "populate_space"},
            "status": trial_module.TrialStatus.RUNNING,
        }

    def score_trial(self, trial_id):
        super().score_trial(trial_id)
        self.score_trial_called = True


def test_hp_not_provided_for_tune_new_entry_error(tmp_path):
    with pytest.raises(ValueError, match="tune_new_entries=False"):
        OracleStub(
            directory=tmp_path, objective="val_loss", tune_new_entries=False
        )


def test_hp_not_provided_for_allow_new_entry_error(tmp_path):
    with pytest.raises(ValueError, match="allow_new_entries=False"):
        OracleStub(
            directory=tmp_path, objective="val_loss", allow_new_entries=False
        )


def test_objective_not_found_error(tmp_path):
    oracle = OracleStub(directory=tmp_path, objective="val_loss")
    trial = oracle.create_trial(tuner_id="a")
    with pytest.raises(ValueError, match="Objective value missing"):
        oracle.update_trial(trial_id=trial.trial_id, metrics={"unknown": 0.5})


def test_multi_objective_found(tmp_path):
    oracle = OracleStub(directory=tmp_path, objective=["val_loss", "mse"])
    trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(
        trial_id=trial.trial_id, metrics={"val_loss": 0.5, "mse": 0.1}
    )


def test_private_populate_space_deprecated_and_call_public(tmp_path):
    oracle = OracleStub(directory=tmp_path, objective="val_loss")
    with pytest.deprecated_call():
        assert isinstance(oracle._populate_space("100"), dict)


def test_private_score_trial_deprecated_and_call_public(tmp_path):
    oracle = OracleStub(directory=tmp_path, objective="val_loss")
    trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(trial_id=trial.trial_id, metrics={"val_loss": 0.5})
    with pytest.deprecated_call():
        oracle._score_trial(trial)
    assert oracle.score_trial_called


def test_import_objective_from_oracle():
    # This test is for backward compatibility.
    from keras_tuner.engine.oracle import Objective

    assert Objective is keras_tuner.Objective


def test_duplicate(tmp_path):
    class MyOracle(OracleStub):
        def populate_space(self, trial_id):
            values = {"hp1": 1}
            if len(self.ongoing_trials) > 0:
                assert self._duplicate(values)
            return {
                "values": values,
                "status": trial_module.TrialStatus.RUNNING,
            }

    oracle = MyOracle(directory=tmp_path, objective="val_loss")
    oracle.create_trial(tuner_id="a")
    oracle.create_trial(tuner_id="b")
    assert len(oracle.ongoing_trials) == 2


def test_end_trial_backward_compatible(tmp_path):
    oracle = OracleStub(directory=tmp_path, objective="val_loss")
    trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(trial.trial_id, {"val_loss": 1.0})
    oracle.end_trial(trial.trial_id, "COMPLETE")


def test_not_duplicate(tmp_path):
    class MyOracle(OracleStub):
        def populate_space(self, trial_id):
            values = {"hp1": len(self.ongoing_trials)}
            assert not self._duplicate(values)
            return {
                "values": values,
                "status": trial_module.TrialStatus.RUNNING,
            }

    oracle = MyOracle(directory=tmp_path, objective="val_loss")
    oracle.create_trial(tuner_id="a")
    oracle.create_trial(tuner_id="b")
    assert len(oracle.ongoing_trials) == 2


def test_new_hp_duplicate(tmp_path):
    class MyOracle(OracleStub):
        def populate_space(self, trial_id):
            values = {"hp1": 1}
            assert not self._duplicate(values)
            if len(self.end_order) > 0:
                values["hp2"] = 2
                assert self._duplicate(values)
            return {
                "values": values,
                "status": trial_module.TrialStatus.RUNNING,
            }

    oracle = MyOracle(directory=tmp_path, objective="val_loss")
    trial = oracle.create_trial(tuner_id="a")
    trial.hyperparameters.values["hp2"] = 2
    oracle.update_trial(trial.trial_id, {"val_loss": 3.0})
    oracle.end_trial(trial)
    oracle.create_trial(tuner_id="b")
    assert len(oracle.start_order) == 2


def test_default_no_retry(tmp_path):
    oracle = OracleStub(directory=tmp_path, objective="val_loss")
    trial_1 = oracle.create_trial(tuner_id="a")
    trial_1.status = trial_module.TrialStatus.INVALID
    trial_1.message = "error1"
    oracle.end_trial(trial_1)

    trial_2 = oracle.create_trial(tuner_id="a")
    assert trial_1.trial_id != trial_2.trial_id


def test_retry_invalid_trial(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    trial_1 = oracle.create_trial(tuner_id="a")
    trial_1.status = trial_module.TrialStatus.INVALID
    trial_1.message = "error1"
    oracle.end_trial(trial_1)

    # This is the retry for the trial.
    trial_2 = oracle.create_trial(tuner_id="a")
    assert trial_1.trial_id == trial_2.trial_id

    # Retried once. This is a new trial.
    trial_3 = oracle.create_trial(tuner_id="b")
    assert trial_1.trial_id != trial_3.trial_id


def test_is_nan_mark_as_invalid(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(trial.trial_id, metrics={"val_loss": float("nan")})
    trial.status = trial_module.TrialStatus.COMPLETED
    trial.message = "error1"
    oracle.end_trial(trial)
    assert (
        oracle.trials[trial.trial_id].status == trial_module.TrialStatus.INVALID
    )


def test_no_retry_for_failed_trial(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    trial_1 = oracle.create_trial(tuner_id="a")
    # Failed, so no retry.
    trial_1.status = trial_module.TrialStatus.FAILED
    trial_1.message = "error1"
    oracle.end_trial(trial_1)

    trial_2 = oracle.create_trial(tuner_id="a")
    assert trial_1.trial_id != trial_2.trial_id


def test_consecutive_failures_in_limit(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=2
    )

    # (1 run + 2 retry) * 2 trial = 6
    for _ in range(6):
        trial = oracle.create_trial(tuner_id="a")
        trial.status = trial_module.TrialStatus.INVALID
        trial.message = "error1"
        oracle.end_trial(trial)

    for _ in range(3):
        trial = oracle.create_trial(tuner_id="a")
        trial.status = trial_module.TrialStatus.COMPLETED
        oracle.update_trial(trial.trial_id, metrics={"val_loss": 0.5})
        oracle.end_trial(trial)


def test_too_many_consecutive_failures(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=2
    )

    with pytest.raises(RuntimeError, match="Number of consecutive") as e:
        for _ in range(3):
            trial = oracle.create_trial(tuner_id="a")
            # Failed, so no retry.
            trial.status = trial_module.TrialStatus.FAILED
            trial.message = "custom_error_info"
            oracle.end_trial(trial)
        assert "custom_error_info" in str(e)


def test_synchronized_functions_in_same_oracle_same_function(tmp_path):
    class MyOracle(OracleStub):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.log = []

        @oracle_module.synchronized
        def create_trial(self, tuner_id):
            # Log ID at the beginning.
            self.log.append(tuner_id)
            time.sleep(0.5)
            # Log ID in the end.
            self.log.append(tuner_id)
            return super().create_trial(tuner_id)

    oracle = MyOracle(directory=tmp_path)

    def thread_function(i):
        oracle.create_trial(tuner_id=str(i))

    threads = []
    for i in range(5):
        thread = threading.Thread(target=thread_function, args=(i,))
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    for i in range(5):
        # The same ID should be next to each other.
        # No other thread interupting between start and end.
        assert oracle.log[i * 2] == oracle.log[i * 2 + 1]


def test_synchronized_functions_in_same_oracle_diff_function(tmp_path):
    class MyOracle(OracleStub):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.log = []

        @oracle_module.synchronized
        def create_trial(self, tuner_id):
            self.log.append("create")
            time.sleep(0.5)
            self.log.append("create")
            return super().create_trial(tuner_id)

        @oracle_module.synchronized
        def end_trial(self, trial):
            self.log.append("end")
            time.sleep(0.5)
            self.log.append("end")
            return super().end_trial(trial)

    oracle = MyOracle(
        directory=tmp_path,
        objective="val_loss",
    )

    trial = oracle.create_trial(tuner_id="a")
    trial.status = trial_module.TrialStatus.COMPLETED
    oracle.update_trial(trial.trial_id, metrics={"val_loss": 0.5})

    def thread_function_create():
        oracle.create_trial(tuner_id="b")

    def thread_function_end():
        oracle.end_trial(trial)

    thread_create = threading.Thread(target=thread_function_create)
    thread_end = threading.Thread(target=thread_function_end)

    thread_create.start()
    thread_end.start()

    thread_create.join()
    thread_end.join()

    for i in range(2):
        # The same ID should be next to each other.
        # No other thread interupting between start and end.
        assert oracle.log[i * 2] == oracle.log[i * 2 + 1]


def test_synchronized_functions_in_different_oracle_doesnt_block(tmp_path):
    log = []

    class MyOracle(OracleStub):
        @oracle_module.synchronized
        def create_trial(self, tuner_id):
            # Log ID at the beginning.
            log.append(tuner_id)
            time.sleep(0.5)
            # Log ID in the end.
            log.append(tuner_id)
            return super().create_trial(tuner_id)

    def thread_function(i):
        oracle = MyOracle(directory=tmp_path)
        oracle.create_trial(tuner_id=str(i))

    threads = []
    for i in range(5):
        thread = threading.Thread(target=thread_function, args=(i,))
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    # All threads begin to sleep before anyone ends.
    assert set(log[:5]) == set(log[5:])


def test_oracle_return_same_trial_if_same_tuner(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    trial_1 = oracle.create_trial(tuner_id="a")
    trial_2 = oracle.create_trial(tuner_id="a")

    assert trial_1.trial_id == trial_2.trial_id


def test_oracle_reload_ongoing_trials_to_retry(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    trial_1 = oracle.create_trial(tuner_id="a")
    trial_2 = oracle.create_trial(tuner_id="b")

    oracle_2 = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    oracle_2.reload()

    trial_3 = oracle.create_trial(tuner_id="a")
    trial_4 = oracle.create_trial(tuner_id="b")

    assert set([trial_3.trial_id, trial_4.trial_id]) == set(
        [trial_1.trial_id, trial_2.trial_id]
    )


def test_get_best_trial_with_nans(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )

    for i in range(10):
        trial = oracle.create_trial(tuner_id="a")
        oracle.update_trial(trial.trial_id, {"val_loss": np.random.rand()})
        trial.status = trial_module.TrialStatus.COMPLETED
        oracle.end_trial(trial)

    best_trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(best_trial.trial_id, {"val_loss": -0.1})
    best_trial.status = trial_module.TrialStatus.COMPLETED
    oracle.end_trial(best_trial)

    trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(trial.trial_id, {"val_loss": float("nan")})
    trial.status = trial_module.TrialStatus.COMPLETED
    oracle.end_trial(trial)

    assert len(oracle.get_best_trials()) > 0
    assert oracle.get_best_trials()[0].trial_id == best_trial.trial_id


def test_overwrite_false_resume(tmp_path):
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    for i in range(10):
        trial = oracle.create_trial(tuner_id="a")
        oracle.update_trial(trial.trial_id, {"val_loss": np.random.rand()})
        trial.status = trial_module.TrialStatus.COMPLETED
        oracle.end_trial(trial)

    trial = oracle.create_trial(tuner_id="a")
    trial_id = trial.trial_id
    oracle = OracleStub(
        directory=tmp_path, objective="val_loss", max_retries_per_trial=1
    )
    oracle.reload()

    trial = oracle.create_trial(tuner_id="a")
    oracle.update_trial(trial.trial_id, {"val_loss": np.random.rand()})
    trial.status = trial_module.TrialStatus.COMPLETED
    oracle.end_trial(trial)

    assert trial.trial_id == trial_id
    assert (
        oracle.get_trial(trial_id).status == trial_module.TrialStatus.COMPLETED
    )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import sys

import pytest

from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import metrics_tracking
from keras_tuner.engine import trial as trial_module


def test_trial_proto():
    hps = hp_module.HyperParameters()
    hps.Int("a", 0, 10, default=3)
    trial = trial_module.Trial(hps, trial_id="trial1", status="COMPLETED")
    trial.metrics.register("score", direction="max")
    trial.metrics.update("score", 10, step=1)

    proto = trial.to_proto()
    assert len(proto.hyperparameters.space.int_space) == 1
    assert proto.hyperparameters.values.values["a"].int_value == 3
    assert not proto.HasField("score")

    new_trial = trial_module.Trial.from_proto(proto)
    assert new_trial.status == trial.status
    assert new_trial.hyperparameters.get("a") == 3
    assert new_trial.trial_id == trial.trial_id
    assert new_trial.score == trial.score
    assert new_trial.best_step == trial.best_step

    trial.score = -10
    trial.best_step = 3

    proto = trial.to_proto()
    assert proto.HasField("score")
    assert proto.score.value == -10
    assert proto.score.step == 3

    new_trial = trial_module.Trial.from_proto(proto)
    assert new_trial.score == -10
    assert new_trial.best_step == 3
    assert new_trial.metrics.get_history("score") == [
        metrics_tracking.MetricObservation(10, step=1)
    ]


def test_trial_status_proto():
    assert (
        trial_module.TrialStatus.from_proto(
            trial_module.TrialStatus.to_proto(None)
        )
        is None
    )
    assert (
        trial_module.TrialStatus.from_proto(
            trial_module.TrialStatus.to_proto(trial_module.TrialStatus.IDLE)
        )
        == trial_module.TrialStatus.IDLE
    )
    assert (
        trial_module.TrialStatus.from_proto(
            trial_module.TrialStatus.to_proto(trial_module.TrialStatus.FAILED)
        )
        == trial_module.TrialStatus.FAILED
    )
    assert (
        trial_module.TrialStatus.from_proto(
            trial_module.TrialStatus.to_proto(trial_module.TrialStatus.INVALID)
        )
        == trial_module.TrialStatus.INVALID
    )
    with pytest.raises(ValueError, match="Unknown status"):
        trial_module.TrialStatus.to_proto("OTHER")
    with pytest.raises(ValueError, match="Unknown status"):
        trial_module.TrialStatus.from_proto(16)


def test_trial_error_in_summary():
    error_message = "stack_trace\nerror_type\n"
    trial = trial_module.Trial(
        hyperparameters=hp_module.HyperParameters(),
        trial_id="3",
        status=trial_module.TrialStatus.FAILED,
        message=error_message,
    )

    stdout = io.StringIO()
    sys.stdout = stdout
    trial.summary()
    sys.stdout = sys.__stdout__
    assert error_message in stdout.getvalue()



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import math

from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import oracle as oracle_module
from keras_tuner.engine import tuner as tuner_module


@keras_tuner_export("keras_tuner.oracles.HyperbandOracle")
class HyperbandOracle(oracle_module.Oracle):
    """Oracle class for Hyperband.

    Note that to use this Oracle with your own subclassed Tuner, your Tuner
    class must be able to handle in `Tuner.run_trial` three special
    hyperparameters that will be set by this Tuner:

    - "tuner/trial_id": String, optionally set. The trial_id of the Trial to
        load from when starting this trial.
    - "tuner/initial_epoch": Int, always set. The initial epoch the Trial should
        be started from.
    - "tuner/epochs": Int, always set. The cumulative number of epochs this
        Trial should be trained.

    These hyperparameters will be set during the "successive halving" portion
    of the Hyperband algorithm.

    Examples:

    ```python
    def run_trial(self, trial, *args, **kwargs):
        hp = trial.hyperparameters
        if "tuner/trial_id" in hp:
            past_trial = self.oracle.get_trial(hp['tuner/trial_id'])
            model = self.load_model(past_trial)
        else:
            model = self.hypermodel.build(hp)

        initial_epoch = hp['tuner/initial_epoch']
        last_epoch = hp['tuner/epochs']

        for epoch in range(initial_epoch, last_epoch):
            self.on_epoch_begin(...)
            for step in range(...):
                # Run model training step here.
            self.on_epoch_end(...)
    ```

    Args:
        objective: A string, `keras_tuner.Objective` instance, or a list of
            `keras_tuner.Objective`s and strings. If a string, the direction of
            the optimization (min or max) will be inferred. If a list of
            `keras_tuner.Objective`, we will minimize the sum of all the
            objectives to minimize subtracting the sum of all the objectives to
            maximize. The `objective` argument is optional when
            `Tuner.run_trial()` or `HyperModel.fit()` returns a single float as
            the objective to minimize.
        max_epochs: Integer, the maximum number of epochs to train one model.
            It is recommended to set this to a value slightly higher than the
            expected epochs to convergence for your largest Model, and to use
            early stopping during training (for example, via
            `tf.keras.callbacks.EarlyStopping`). Defaults to 100.
        factor: Integer, the reduction factor for the number of epochs
            and number of models for each bracket. Defaults to 3.
        hyperband_iterations: Integer, at least 1, the number of times to
            iterate over the full Hyperband algorithm. One iteration will run
            approximately `max_epochs * (math.log(max_epochs, factor) ** 2)`
            cumulative epochs across all trials. It is recommended to set this
            to as high a value as is within your resource budget. Defaults to
            1.
        seed: Optional integer, the random seed.
        hyperparameters: Optional HyperParameters instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        max_retries_per_trial: Integer. Defaults to 0. The maximum number of
            times to retry a `Trial` if the trial crashed or the results are
            invalid.
        max_consecutive_failed_trials: Integer. Defaults to 3. The maximum
            number of consecutive failed `Trial`s. When this number is reached,
            the search will be stopped. A `Trial` is marked as failed when none
            of the retries succeeded.
    """

    def __init__(
        self,
        objective=None,
        max_epochs=100,
        factor=3,
        hyperband_iterations=1,
        seed=None,
        hyperparameters=None,
        allow_new_entries=True,
        tune_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
    ):
        super().__init__(
            objective=objective,
            hyperparameters=hyperparameters,
            allow_new_entries=allow_new_entries,
            tune_new_entries=tune_new_entries,
            seed=seed,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        if factor < 2:
            raise ValueError("factor needs to be a int larger than 1.")

        self.hyperband_iterations = hyperband_iterations or float("inf")
        self.max_epochs = max_epochs
        # Minimum epochs before successive halving, Hyperband sweeps through
        # varying degress of aggressiveness.
        self.min_epochs = 1
        self.factor = factor

        self._current_iteration = 0
        # Start with most aggressively halving bracket.
        self._current_bracket = self._get_num_brackets() - 1
        self._brackets = []

        self._start_new_bracket()

    def populate_space(self, trial_id):
        """Fill the hyperparameter space with values.

        Args:
            trial_id: A string, the ID for this Trial.

        Returns:
            A dictionary with keys "values" and "status", where "values" is
            a mapping of parameter names to suggested values, and "status"
            should be one of "RUNNING" (the trial can start normally), "IDLE"
            (the oracle is waiting on something and cannot create a trial), or
            "STOPPED" (the oracle has finished searching and no new trial should
            be created).
        """
        self._remove_completed_brackets()

        for bracket in self._brackets:
            bracket_num = bracket["bracket_num"]
            rounds = bracket["rounds"]

            if len(rounds[0]) < self._get_size(bracket_num, round_num=0):
                # Populate the initial random trials for this bracket.
                return self._random_trial(trial_id, bracket)
            # Try to populate incomplete rounds for this bracket.
            for round_num in range(1, len(rounds)):
                round_info = rounds[round_num]
                past_round_info = rounds[round_num - 1]
                size = self._get_size(bracket_num, round_num)
                past_size = self._get_size(bracket_num, round_num - 1)

                # If more trials from the last round are ready than will be
                # thrown out, we can select the best to run for the next round.
                already_selected = [info["past_id"] for info in round_info]
                candidates = [
                    self.trials[info["id"]]
                    for info in past_round_info
                    if info["id"] not in already_selected
                ]
                candidates = [t for t in candidates if t.status == "COMPLETED"]
                if len(candidates) > past_size - size:
                    sorted_candidates = sorted(
                        candidates,
                        key=lambda t: t.score,
                        reverse=self.objective.direction == "max",
                    )
                    best_trial = sorted_candidates[0]

                    values = best_trial.hyperparameters.values.copy()
                    values["tuner/trial_id"] = best_trial.trial_id
                    values["tuner/epochs"] = self._get_epochs(
                        bracket_num, round_num
                    )
                    values["tuner/initial_epoch"] = self._get_epochs(
                        bracket_num, round_num - 1
                    )
                    values["tuner/bracket"] = self._current_bracket
                    values["tuner/round"] = round_num

                    round_info.append(
                        {"past_id": best_trial.trial_id, "id": trial_id}
                    )
                    return {"status": "RUNNING", "values": values}

        # This is reached if no trials from current brackets can be run.

        # Max sweeps has been reached, no more brackets should be created.
        if (
            self._current_bracket == 0
            and self._current_iteration + 1 == self.hyperband_iterations
        ):
            # Stop creating new brackets, but wait to complete other brackets.
            if self.ongoing_trials:
                return {"status": "IDLE"}
            return {"status": "STOPPED"}
        # Create a new bracket.
        else:
            self._increment_bracket_num()
            self._start_new_bracket()
            return self._random_trial(trial_id, self._brackets[-1])

    def _start_new_bracket(self):
        rounds = []
        rounds.extend(
            [] for _ in range(self._get_num_rounds(self._current_bracket))
        )
        bracket = {"bracket_num": self._current_bracket, "rounds": rounds}
        self._brackets.append(bracket)

    def _increment_bracket_num(self):
        self._current_bracket -= 1
        if self._current_bracket < 0:
            self._current_bracket = self._get_num_brackets() - 1
            self._current_iteration += 1

    def _remove_completed_brackets(self):
        # Filter out completed brackets.
        def _bracket_is_incomplete(bracket):
            bracket_num = bracket["bracket_num"]
            rounds = bracket["rounds"]
            last_round = len(rounds) - 1
            return len(rounds[last_round]) != self._get_size(
                bracket_num, last_round
            )

        self._brackets = list(filter(_bracket_is_incomplete, self._brackets))

    def _compute_values_hash(self, values):
        values = copy.copy(values)
        values.pop("tuner/epochs", None)
        values.pop("tuner/initial_epoch", None)
        values.pop("tuner/bracket", None)
        values.pop("tuner/round", None)
        return super()._compute_values_hash(values)

    def _random_trial(self, trial_id, bracket):
        bracket_num = bracket["bracket_num"]
        rounds = bracket["rounds"]
        values = self._random_values()
        if values:
            values["tuner/epochs"] = self._get_epochs(bracket_num, 0)
            values["tuner/initial_epoch"] = 0
            values["tuner/bracket"] = self._current_bracket
            values["tuner/round"] = 0
            rounds[0].append({"past_id": None, "id": trial_id})
            return {"status": "RUNNING", "values": values}
        elif self.ongoing_trials:
            # Can't create new random values, but successive halvings may still
            # be needed.
            return {"status": "IDLE"}
        else:
            # Collision and no ongoing trials should trigger an exit.
            return {"status": "STOPPED"}

    def _get_size(self, bracket_num, round_num):
        # Set up so that each bracket takes approx. the same amount of
        # resources.
        bracket0_end_size = math.ceil(
            1 + math.log(self.max_epochs, self.factor)
        )
        bracket_end_size = bracket0_end_size / (bracket_num + 1)
        return math.ceil(
            bracket_end_size * self.factor ** (bracket_num - round_num)
        )

    def _get_epochs(self, bracket_num, round_num):
        return math.ceil(
            self.max_epochs / self.factor ** (bracket_num - round_num)
        )

    def _get_num_rounds(self, bracket_num):
        # Bracket 0 just runs random search, others do successive halving.
        return bracket_num + 1

    def _get_num_brackets(self):
        epochs = self.max_epochs
        brackets = 0
        while epochs >= self.min_epochs:
            epochs = epochs / self.factor
            brackets += 1
        return brackets

    def get_state(self):
        state = super().get_state()
        state.update(
            {
                "hyperband_iterations": self.hyperband_iterations,
                "max_epochs": self.max_epochs,
                "min_epochs": self.min_epochs,
                "factor": self.factor,
                "brackets": self._brackets,
                "current_bracket": self._current_bracket,
                "current_iteration": self._current_iteration,
            }
        )
        return state

    def set_state(self, state):
        super().set_state(state)
        self.hyperband_iterations = state["hyperband_iterations"]
        self.max_epochs = state["max_epochs"]
        self.min_epochs = state["min_epochs"]
        self.factor = state["factor"]
        self._brackets = state["brackets"]
        self._current_bracket = state["current_bracket"]
        self._current_iteration = state["current_iteration"]


@keras_tuner_export(["keras_tuner.Hyperband", "keras_tuner.tuners.Hyperband"])
class Hyperband(tuner_module.Tuner):
    """Variation of HyperBand algorithm.

    Reference:
        Li, Lisha, and Kevin Jamieson.
        ["Hyperband: A Novel Bandit-Based
         Approach to Hyperparameter Optimization."
        Journal of Machine Learning Research 18 (2018): 1-52](
            http://jmlr.org/papers/v18/16-558.html).


    Args:
        hypermodel: Instance of `HyperModel` class (or callable that takes
            hyperparameters and returns a `Model` instance). It is optional
            when `Tuner.run_trial()` is overriden and does not use
            `self.hypermodel`.
        objective: A string, `keras_tuner.Objective` instance, or a list of
            `keras_tuner.Objective`s and strings. If a string, the direction of
            the optimization (min or max) will be inferred. If a list of
            `keras_tuner.Objective`, we will minimize the sum of all the
            objectives to minimize subtracting the sum of all the objectives to
            maximize. The `objective` argument is optional when
            `Tuner.run_trial()` or `HyperModel.fit()` returns a single float as
            the objective to minimize.
        max_epochs: Integer, the maximum number of epochs to train one model.
            It is recommended to set this to a value slightly higher than the
            expected epochs to convergence for your largest Model, and to use
            early stopping during training (for example, via
            `tf.keras.callbacks.EarlyStopping`). Defaults to 100.
        factor: Integer, the reduction factor for the number of epochs
            and number of models for each bracket. Defaults to 3.
        hyperband_iterations: Integer, at least 1, the number of times to
            iterate over the full Hyperband algorithm. One iteration will run
            approximately `max_epochs * (math.log(max_epochs, factor) ** 2)`
            cumulative epochs across all trials. It is recommended to set this
            to as high a value as is within your resource budget. Defaults to
            1.
        seed: Optional integer, the random seed.
        hyperparameters: Optional HyperParameters instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        max_retries_per_trial: Integer. Defaults to 0. The maximum number of
            times to retry a `Trial` if the trial crashed or the results are
            invalid.
        max_consecutive_failed_trials: Integer. Defaults to 3. The maximum
            number of consecutive failed `Trial`s. When this number is reached,
            the search will be stopped. A `Trial` is marked as failed when none
            of the retries succeeded.
        **kwargs: Keyword arguments relevant to all `Tuner` subclasses.
            Please see the docstring for `Tuner`.
    """

    def __init__(
        self,
        hypermodel=None,
        objective=None,
        max_epochs=100,
        factor=3,
        hyperband_iterations=1,
        seed=None,
        hyperparameters=None,
        tune_new_entries=True,
        allow_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
        **kwargs
    ):
        oracle = HyperbandOracle(
            objective,
            max_epochs=max_epochs,
            factor=factor,
            hyperband_iterations=hyperband_iterations,
            seed=seed,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        super().__init__(oracle=oracle, hypermodel=hypermodel, **kwargs)

    def run_trial(self, trial, *fit_args, **fit_kwargs):
        hp = trial.hyperparameters
        if "tuner/epochs" in hp.values:
            fit_kwargs["epochs"] = hp.values["tuner/epochs"]
            fit_kwargs["initial_epoch"] = hp.values["tuner/initial_epoch"]
        return super().run_trial(trial, *fit_args, **fit_kwargs)

    def _build_hypermodel(self, hp):
        model = super()._build_hypermodel(hp)
        if "tuner/trial_id" in hp.values:
            trial_id = hp.values["tuner/trial_id"]
            # Load best checkpoint from this trial.
            model.load_weights(self._get_checkpoint_fname(trial_id))
        return model



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for HyperImageAugment Models."""

import os

import numpy as np
import pytest
import tensorflow as tf
from packaging.version import parse
from tensorflow import keras

from keras_tuner.applications import augment as aug_module
from keras_tuner.engine import hyperparameters as hp_module


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Causes CI to stall")
@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_transforms_search_space():
    hm = aug_module.HyperImageAugment(input_shape=(32, 32, 3))
    # Default choice
    assert hm.transforms == [
        ("rotate", (0, 0.5)),
        ("translate_x", (0, 0.4)),
        ("translate_y", (0, 0.4)),
        ("contrast", (0, 0.3)),
    ]

    hm = aug_module.HyperImageAugment(
        input_shape=(32, 32, 3),
        rotate=0.3,
        translate_x=[0.1, 0.5],
        contrast=None,
    )
    assert hm.transforms == [
        ("rotate", (0, 0.3)),
        ("translate_x", (0.1, 0.5)),
        ("translate_y", (0, 0.4)),
    ]


def test_tf_version_too_low_error():
    pp_module = aug_module.preprocessing
    aug_module.preprocessing = None

    with pytest.raises(ImportError, match="HyperImageAugment requires"):
        aug_module.HyperImageAugment()

    aug_module.preprocessing = pp_module


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_input_requirement():
    hp = hp_module.HyperParameters()
    with pytest.raises(ValueError, match=r".*must specify.*"):
        hm = aug_module.HyperImageAugment()

    hm = aug_module.HyperImageAugment(input_shape=(None, None, 3))
    model = hm.build(hp)
    assert model.built

    hm = aug_module.HyperImageAugment(
        input_tensor=keras.Input(shape=(32, 32, 3))
    )
    model = hm.build(hp)
    assert model.built


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_model_construction_factor_zero():
    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(input_shape=(None, None, 3))
    model = hm.build(hp)
    # augment_layers search default space [0, 4], with default zero.
    assert len(model.layers) == 1

    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(
        input_shape=(None, None, 3), augment_layers=0
    )
    model = hm.build(hp)
    # factors default all zero, the model should only have input layer
    assert len(model.layers) == 1


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_model_construction_fixed_aug():
    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(
        input_shape=(None, None, 3), rotate=[0.2, 0.5], augment_layers=0
    )
    model = hm.build(hp)
    assert model.layers
    assert model.name == "image_augment"

    # Output shape includes batch dimension.
    assert model.output_shape == (None, None, None, 3)
    out = model.predict(np.ones((1, 32, 32, 3)))
    assert out.shape == (1, 32, 32, 3)
    # Augment does not distort image when inferencing.
    assert (out != 1).sum() == 0


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_model_construction_rand_aug():
    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(
        input_shape=(None, None, 3), rotate=[0.2, 0.5]
    )
    model = hm.build(hp)
    assert model.layers
    assert model.name == "image_rand_augment"

    # Output shape includes batch dimension.
    assert model.output_shape == (None, None, None, 3)
    out = model.predict(np.ones((1, 32, 32, 3)))
    assert out.shape == (1, 32, 32, 3)
    # Augment does not distort image when inferencing.
    assert (out != 1).sum() == 0


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_hyperparameter_selection_and_hp_defaults_fixed_aug():
    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(
        input_shape=(32, 32, 3),
        translate_x=[0.2, 0.4],
        contrast=None,
        augment_layers=0,
    )
    hm.build(hp)
    # default value of default search space are always minimum.
    assert hp.get("factor_rotate") == 0
    assert hp.get("factor_translate_x") == 0.2
    assert hp.get("factor_translate_y") == 0
    assert "factor_contrast" not in hp.values


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_hyperparameter_existence_and_hp_defaults_rand_aug():
    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(
        input_shape=(32, 32, 3), augment_layers=[2, 5], contrast=False
    )
    hm.build(hp)
    assert hp.get("augment_layers") == 2


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_augment_layers_not_int():
    with pytest.raises(ValueError, match="must be int"):
        hp = hp_module.HyperParameters()
        hm = aug_module.HyperImageAugment(
            input_shape=(32, 32, 3), augment_layers=1.5
        )
        hm.build(hp)


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_contrast_0_5_to_1():
    hp = hp_module.HyperParameters()
    hm = aug_module.HyperImageAugment(
        input_shape=(32, 32, 3), augment_layers=[1, 2], contrast=[0.5, 1]
    )
    hm.build(hp)


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_3_range_params_raise_error():
    with pytest.raises(ValueError, match="exceed 2"):
        hp = hp_module.HyperParameters()
        hm = aug_module.HyperImageAugment(
            input_shape=(32, 32, 3), augment_layers=[1, 2], contrast=[0.5, 1, 2]
        )
        hm.build(hp)


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_contrast_range_receive_string_raise_error():
    with pytest.raises(ValueError, match="must be int or float"):
        hp = hp_module.HyperParameters()
        hm = aug_module.HyperImageAugment(
            input_shape=(32, 32, 3), augment_layers=[1, 2], contrast=[0.5, "a"]
        )
        hm.build(hp)


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_hyperparameter_override_fixed_aug():
    hp = hp_module.HyperParameters()
    hp.Fixed("factor_rotate", 0.9)
    hp.Choice("factor_translate_x", [0.8])
    hm = aug_module.HyperImageAugment(input_shape=(32, 32, 3), augment_layers=0)
    hm.build(hp)
    assert hp.get("factor_rotate") == 0.9
    assert hp.get("factor_translate_x") == 0.8
    assert hp.get("factor_translate_y") == 0.0
    assert hp.get("factor_contrast") == 0.0


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers only exist in TF2.3+.",
)
def test_hyperparameter_override_rand_aug():
    hp = hp_module.HyperParameters()
    hp.Fixed("randaug_mag", 1.0)
    hp.Choice("randaug_count", [4])
    hm = aug_module.HyperImageAugment(
        input_shape=(32, 32, 3), augment_layers=[2, 4]
    )
    hm.build(hp)
    assert hp.get("randaug_mag") == 1.0
    assert hp.get("randaug_count") == 4



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"HyperParameters logic."

import collections
import contextlib
import copy

import six

from keras_tuner import protos
from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import conditions as conditions_mod
from keras_tuner.engine.hyperparameters import hp_types
from keras_tuner.engine.hyperparameters import hyperparameter as hp_module


@keras_tuner_export("keras_tuner.HyperParameters")
class HyperParameters:
    """Container for both a hyperparameter space, and current values.

    A `HyperParameters` instance can be pass to `HyperModel.build(hp)` as an
    argument to build a model.

    To prevent the users from depending on inactive hyperparameter values, only
    active hyperparameters should have values in `HyperParameters.values`.

    Attributes:
        space: A list of `HyperParameter` objects.
        values: A dict mapping hyperparameter names to current values.
    """

    def __init__(self):
        # Current name scopes.
        self._name_scopes = []
        # Current `Condition`s, managed by `conditional_scope`.
        self._conditions = []

        # Dict of list of hyperparameters with same
        # name but different conditions, e.g. `{name: [hp1, hp2]}`.
        # Hyperparameters are uniquely identified by their name and
        # conditions.
        self._hps = collections.defaultdict(list)

        # List of hyperparameters, maintained in insertion order.
        # This guarantees that conditional params are always later in
        # the list than their parents.
        self._space = []

        # Active values for this `Trial`.
        self.values = {}

        # A list of active `conditional_scope`s in a build,
        # each of which is a list of condtions.
        # Used by BaseTuner to activate all conditions.
        # No need to empty these after builds since when building the model, hp
        # is copied from the Oracle's hp, which always have these 2 fields
        # empty.
        self.active_scopes = []
        # Similar for inactive `conditional_scope`s.
        self.inactive_scopes = []

    @contextlib.contextmanager
    def name_scope(self, name):
        self._name_scopes.append(name)
        try:
            yield
        finally:
            self._name_scopes.pop()

    @contextlib.contextmanager
    def conditional_scope(self, parent_name, parent_values):
        """Opens a scope to create conditional HyperParameters.

        All `HyperParameter`s created under this scope will only be active when
        the parent `HyperParameter` specified by `parent_name` is equal to one
        of the values passed in `parent_values`.

        When the condition is not met, creating a `HyperParameter` under this
        scope will register the `HyperParameter`, but will return `None` rather
        than a concrete value.

        Note that any Python code under this scope will execute regardless of
        whether the condition is met.

        This feature is for the `Tuner` to collect more information of the
        search space and the current trial.  It is especially useful for model
        selection. If the parent `HyperParameter` is for model selection, the
        `HyperParameter`s in a model should only be active when the model
        selected, which can be implemented using `conditional_scope`.

        Examples:

        ```python
        def MyHyperModel(HyperModel):
            def build(self, hp):
                model = Sequential()
                model.add(Input(shape=(32, 32, 3)))
                model_type = hp.Choice("model_type", ["mlp", "cnn"])
                with hp.conditional_scope("model_type", ["mlp"]):
                    if model_type == "mlp":
                        model.add(Flatten())
                        model.add(Dense(32, activation='relu'))
                with hp.conditional_scope("model_type", ["cnn"]):
                    if model_type == "cnn":
                        model.add(Conv2D(64, 3, activation='relu'))
                        model.add(GlobalAveragePooling2D())
                model.add(Dense(10, activation='softmax'))
                return model
        ```

        Args:
            parent_name: A string, specifying the name of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.
            parent_values: A list of the values of the parent `HyperParameter`
                to use as the condition to activate the current
                `HyperParameter`.
        """
        parent_name = self._get_name(parent_name)  # Add name_scopes.
        if not self._exists(parent_name):
            raise ValueError(
                f"`HyperParameter` named: {parent_name} not defined."
            )

        condition = conditions_mod.Parent(parent_name, parent_values)
        self._conditions.append(condition)

        if condition.is_active(self.values):
            self.active_scopes.append(copy.deepcopy(self._conditions))
        else:
            self.inactive_scopes.append(copy.deepcopy(self._conditions))

        try:
            yield
        finally:
            self._conditions.pop()

    def is_active(self, hyperparameter):
        """Checks if a hyperparameter is currently active for a `Trial`.

        A hyperparameter is considered active if and only if all its parent
        conditions are active, and not affected by whether the hyperparameter
        is used while building the model. The function is usually called by the
        `Oracle` for populating new hyperparameter values and updating the trial
        after receiving the evaluation results.

        Args:
            hp: A string or `HyperParameter` instance. If string, checks whether
                any hyperparameter with that name is active. If `HyperParameter`
                instance, checks whether the object is active.

        Returns:
            A boolean, whether the hyperparameter is active.
        """
        hp = hyperparameter
        if isinstance(hp, hp_module.HyperParameter):
            return self._conditions_are_active(hp.conditions)
        hp_name = str(hp)
        return any(
            self._conditions_are_active(temp_hp.conditions)
            for temp_hp in self._hps[hp_name]
        )

    def _conditions_are_active(self, conditions):
        return all(condition.is_active(self.values) for condition in conditions)

    def _exists(self, name, conditions=None):
        """Checks for a hyperparameter with the same name and conditions."""
        if conditions is None:
            conditions = self._conditions

        if name in self._hps:
            hps = self._hps[name]
            for hp in hps:
                if hp.conditions == conditions:
                    return True
        return False

    def _retrieve(self, hp):
        """Gets or creates a hyperparameter.

        Args:
            hp: A `HyperParameter` instance.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        if self._exists(hp.name, hp.conditions):
            if self.is_active(hp):
                return self.values[hp.name]
            return None  # Ensures inactive values are not relied on by user.
        return self._register(hp)

    def _register(self, hyperparameter, overwrite=False):
        """Registers a hyperparameter in this container.

        Args:
            hp: A `HyperParameter` instance.
            overwrite: Boolean, whether to overwrite the existing value with
                the default hyperparameter value.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        hp = hyperparameter
        self._validate_name(hp.name)
        # Copy to ensure this param can be serialized.
        hp = hp.__class__.from_config(hp.get_config())
        self._hps[hp.name].append(hp)
        self._space.append(hp)
        # Only add active values to `self.values`.
        if self.is_active(hp):
            # Use the default value only if not populated.
            if overwrite or hp.name not in self.values:
                self.values[hp.name] = hp.default
            return self.values[hp.name]
        return None  # Ensures inactive values are not relied on by user.

    def get(self, name):
        """Return the current value of this hyperparameter set."""
        name = self._get_name(name)  # Add name_scopes.
        if name in self.values:
            return self.values[name]  # Only active values are added here.
        elif name in self._hps:
            raise ValueError(f"{name} is currently inactive.")
        else:
            raise KeyError(f"{name} does not exist.")

    def __getitem__(self, name):
        return self.get(name)

    def __contains__(self, name):
        try:
            self.get(name)
            return True
        except (KeyError, ValueError):
            return False

    def Choice(
        self,
        name,
        values,
        ordered=None,
        default=None,
        parent_name=None,
        parent_values=None,
    ):
        """Choice of one value among a predefined set of possible values.

        Args:
            name: A string. the name of parameter. Must be unique for each
                `HyperParameter` instance in the search space.
            values: A list of possible values. Values must be int, float,
                str, or bool. All values must be of the same type.
            ordered: Optional boolean, whether the values passed should be
                considered to have an ordering. Defaults to `True` for float/int
                values.  Must be `False` for any other values.
            default: Optional default value to return for the parameter.
                If unspecified, the default value will be:
                - None if None is one of the choices in `values`
                - The first entry in `values` otherwise.
            parent_name: Optional string, specifying the name of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.
            parent_values: Optional list of the values of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        with self._maybe_conditional_scope(parent_name, parent_values):
            hp = hp_types.Choice(
                name=self._get_name(name),  # Add name_scopes.
                values=values,
                ordered=ordered,
                default=default,
                conditions=self._conditions,
            )
            return self._retrieve(hp)

    def Int(
        self,
        name,
        min_value,
        max_value,
        step=None,
        sampling="linear",
        default=None,
        parent_name=None,
        parent_values=None,
    ):
        """Integer hyperparameter.

        Note that unlike Python's `range` function, `max_value` is *included* in
        the possible values this parameter can take on.


        Example #1:

        ```py
        hp.Int(
            "n_layers",
            min_value=6,
            max_value=12)
        ```

        The possible values are [6, 7, 8, 9, 10, 11, 12].

        Example #2:

        ```py
        hp.Int(
            "n_layers",
            min_value=6,
            max_value=13,
            step=3)
        ```

        `step` is the minimum distance between samples.
        The possible values are [6, 9, 12].

        Example #3:

        ```py
        hp.Int(
            "batch_size",
            min_value=2,
            max_value=32,
            step=2,
            sampling="log")
        ```

        When `sampling="log"` the `step` is multiplied between samples.
        The possible values are [2, 4, 8, 16, 32].

        Args:
            name: A string. the name of parameter. Must be unique for each
                `HyperParameter` instance in the search space.
            min_value: Integer, the lower limit of range, inclusive.
            max_value: Integer, the upper limit of range, inclusive.
            step: Optional integer, the distance between two consecutive samples
                in the range. If left unspecified, it is possible to sample any
                integers in the interval. If `sampling="linear"`, it will be the
                minimum additve between two samples. If `sampling="log"`, it
                will be the minimum multiplier between two samples.
            sampling: String. One of "linear", "log", "reverse_log". Defaults to
                "linear". When sampling value, it always start from a value in
                range [0.0, 1.0). The `sampling` argument decides how the value
                is projected into the range of [min_value, max_value].
                "linear": min_value + value * (max_value - min_value)
                "log": min_value * (max_value / min_value) ^ value
                "reverse_log":
                    (max_value -
                     min_value * ((max_value / min_value) ^ (1 - value) - 1))
            default: Integer, default value to return for the parameter. If
                unspecified, the default value will be `min_value`.
            parent_name: Optional string, specifying the name of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.
            parent_values: Optional list of the values of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        with self._maybe_conditional_scope(parent_name, parent_values):
            hp = hp_types.Int(
                name=self._get_name(name),  # Add name_scopes.
                min_value=min_value,
                max_value=max_value,
                step=step,
                sampling=sampling,
                default=default,
                conditions=self._conditions,
            )
            return self._retrieve(hp)

    def Float(
        self,
        name,
        min_value,
        max_value,
        step=None,
        sampling="linear",
        default=None,
        parent_name=None,
        parent_values=None,
    ):
        """Floating point value hyperparameter.

        Example #1:

        ```py
        hp.Float(
            "image_rotation_factor",
            min_value=0,
            max_value=1)
        ```

        All values in interval [0, 1] have equal probability of being sampled.

        Example #2:

        ```py
        hp.Float(
            "image_rotation_factor",
            min_value=0,
            max_value=1,
            step=0.2)
        ```

        `step` is the minimum distance between samples.
        The possible values are [0, 0.2, 0.4, 0.6, 0.8, 1.0].

        Example #3:

        ```py
        hp.Float(
            "learning_rate",
            min_value=0.001,
            max_value=10,
            step=10,
            sampling="log")
        ```

        When `sampling="log"`, the `step` is multiplied between samples.
        The possible values are [0.001, 0.01, 0.1, 1, 10].

        Args:
            name: A string. the name of parameter. Must be unique for each
                `HyperParameter` instance in the search space.
            min_value: Float, the lower bound of the range.
            max_value: Float, the upper bound of the range.
            step: Optional float, the distance between two consecutive samples
                in the range. If left unspecified, it is possible to sample any
                value in the interval. If `sampling="linear"`, it will be the
                minimum additve between two samples. If `sampling="log"`, it
                will be the minimum multiplier between two samples.
            sampling: String. One of "linear", "log", "reverse_log". Defaults to
                "linear". When sampling value, it always start from a value in
                range [0.0, 1.0). The `sampling` argument decides how the value
                is projected into the range of [min_value, max_value].
                "linear": min_value + value * (max_value - min_value)
                "log": min_value * (max_value / min_value) ^ value
                "reverse_log":
                    (max_value -
                     min_value * ((max_value / min_value) ^ (1 - value) - 1))
            default: Float, the default value to return for the parameter. If
                unspecified, the default value will be `min_value`.
            parent_name: Optional string, specifying the name of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.
            parent_values: Optional list of the values of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        with self._maybe_conditional_scope(parent_name, parent_values):
            hp = hp_types.Float(
                name=self._get_name(name),  # Add name_scopes.
                min_value=min_value,
                max_value=max_value,
                step=step,
                sampling=sampling,
                default=default,
                conditions=self._conditions,
            )
            return self._retrieve(hp)

    def Boolean(
        self, name, default=False, parent_name=None, parent_values=None
    ):
        """Choice between True and False.

        Args:
            name: A string. the name of parameter. Must be unique for each
                `HyperParameter` instance in the search space.
            default: Boolean, the default value to return for the parameter.
                If unspecified, the default value will be False.
            parent_name: Optional string, specifying the name of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.
            parent_values: Optional list of the values of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        with self._maybe_conditional_scope(parent_name, parent_values):
            hp = hp_types.Boolean(
                name=self._get_name(name),  # Add name_scopes.
                default=default,
                conditions=self._conditions,
            )
            return self._retrieve(hp)

    def Fixed(self, name, value, parent_name=None, parent_values=None):
        """Fixed, untunable value.

        Args:
            name: A string. the name of parameter. Must be unique for each
                `HyperParameter` instance in the search space.
            value: The value to use (can be any JSON-serializable Python type).
            parent_name: Optional string, specifying the name of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.
            parent_values: Optional list of the values of the parent
                `HyperParameter` to use as the condition to activate the
                current `HyperParameter`.

        Returns:
            The value of the hyperparameter, or None if the hyperparameter is
            not active.
        """
        with self._maybe_conditional_scope(parent_name, parent_values):
            hp = hp_types.Fixed(
                name=self._get_name(name),  # Add name_scopes.
                value=value,
                conditions=self._conditions,
            )
            return self._retrieve(hp)

    @property
    def space(self):
        return self._space

    def get_config(self):
        return {
            "space": [
                {"class_name": p.__class__.__name__, "config": p.get_config()}
                for p in self.space
            ],
            "values": dict(self.values.items()),
        }

    @classmethod
    def from_config(cls, config):
        hps = cls()
        for p in config["space"]:
            p = hp_types.deserialize(p)
            hps._hps[p.name].append(p)
            hps._space.append(p)
        hps.values = dict(config["values"].items())
        return hps

    def copy(self):
        return HyperParameters.from_config(self.get_config())

    def merge(self, hps, overwrite=True):
        """Merges hyperparameters into this object.

        Args:
            hps: A `HyperParameters` object or list of `HyperParameter`
                objects.
            overwrite: Boolean, whether existing `HyperParameter`s should be
                overridden by those in `hps` with the same name and conditions.
        """
        if isinstance(hps, HyperParameters):
            hps = hps.space

        if not overwrite:
            hps = [hp for hp in hps if not self._exists(hp.name, hp.conditions)]

        for hp in hps:
            self._register(hp, overwrite)

    def ensure_active_values(self):
        """Add and remove values if necessary.

        KerasTuner requires only active values to be populated. This function
        removes the inactive values and add the missing active values.

        Args:
            hps: HyperParameters, whose values to be ensured.
        """
        for hp in self.space:
            if self.is_active(hp):
                if hp.name not in self.values:
                    self.values[hp.name] = hp.random_sample()
            else:
                self.values.pop(hp.name, None)

    @classmethod
    def from_proto(cls, proto):
        hps = cls()

        space = []
        if isinstance(proto, protos.get_proto().HyperParameters.Values):
            # Allows passing in only values, space becomes `Fixed`.
            space.extend(
                hp_types.Fixed(name, getattr(value, value.WhichOneof("kind")))
                for name, value in proto.values.items()
            )

        else:
            space.extend(
                hp_types.Fixed.from_proto(fixed_proto)
                for fixed_proto in proto.space.fixed_space
            )

            space.extend(
                hp_types.Float.from_proto(float_proto)
                for float_proto in proto.space.float_space
            )

            space.extend(
                hp_types.Int.from_proto(int_proto)
                for int_proto in proto.space.int_space
            )

            space.extend(
                hp_types.Choice.from_proto(choice_proto)
                for choice_proto in proto.space.choice_space
            )

            space.extend(
                hp_types.Boolean.from_proto(boolean_proto)
                for boolean_proto in proto.space.boolean_space
            )

        hps.merge(space)

        if isinstance(proto, protos.get_proto().HyperParameters.Values):
            values = proto.values
        else:
            values = proto.values.values
        for name, val in values.items():
            hps.values[name] = getattr(val, val.WhichOneof("kind"))

        return hps

    def to_proto(self):
        fixed_space = []
        float_space = []
        int_space = []
        choice_space = []
        boolean_space = []
        for hp in self.space:
            if isinstance(hp, hp_types.Fixed):
                fixed_space.append(hp.to_proto())
            elif isinstance(hp, hp_types.Float):
                float_space.append(hp.to_proto())
            elif isinstance(hp, hp_types.Int):
                int_space.append(hp.to_proto())
            elif isinstance(hp, hp_types.Choice):
                choice_space.append(hp.to_proto())
            elif isinstance(hp, hp_types.Boolean):
                boolean_space.append(hp.to_proto())
            else:
                raise ValueError(f"Unrecognized HP type: {hp}")

        values = {}
        for name, value in self.values.items():
            if isinstance(value, bool):
                val = protos.get_proto().Value(boolean_value=value)
            elif isinstance(value, float):
                val = protos.get_proto().Value(float_value=value)
            elif isinstance(value, six.integer_types):
                val = protos.get_proto().Value(int_value=value)
            elif isinstance(value, six.string_types):
                val = protos.get_proto().Value(string_value=value)
            else:
                raise ValueError(f"Unrecognized value type: {value}")
            values[name] = val

        return protos.get_proto().HyperParameters(
            space=protos.get_proto().HyperParameters.Space(
                fixed_space=fixed_space,
                float_space=float_space,
                int_space=int_space,
                choice_space=choice_space,
                boolean_space=boolean_space,
            ),
            values=protos.get_proto().HyperParameters.Values(values=values),
        )

    @contextlib.contextmanager
    def _maybe_conditional_scope(self, parent_name, parent_values):
        if parent_name:
            with self.conditional_scope(parent_name, parent_values):
                yield
        else:
            yield

    def _get_name(self, name, name_scopes=None):
        """Returns a name qualified by `name_scopes`."""
        if name_scopes is None:
            name_scopes = self._name_scopes

        return (
            "/".join(name_scopes) + "/" + str(name)
            if name_scopes
            else str(name)
        )

    def _validate_name(self, name):
        for condition in self._conditions:
            if condition.name == name:
                raise ValueError(
                    "A conditional `HyperParameter` cannot have the same "
                    "name as its parent. Found: " + str(name) + " and "
                    "parent_name: " + str(condition.name)
                )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from keras_tuner.engine import hyperparameters as hp_module


def test_fixed():
    fixed = hp_module.Fixed("fixed", "value")
    fixed = hp_module.Fixed.from_config(fixed.get_config())
    assert fixed.default == "value"
    assert fixed.random_sample() == "value"

    fixed = hp_module.Fixed("fixed", True)
    assert fixed.default is True
    assert fixed.random_sample() is True

    fixed = hp_module.Fixed("fixed", False)
    fixed = hp_module.Fixed.from_config(fixed.get_config())
    assert fixed.default is False
    assert fixed.random_sample() is False

    fixed = hp_module.Fixed("fixed", 1)
    assert fixed.value == 1
    assert fixed.random_sample() == 1

    fixed = hp_module.Fixed("fixed", 8.2)
    assert fixed.value == 8.2
    assert fixed.random_sample() == 8.2
    assert fixed.value_to_prob(fixed.value) == 0.5

    with pytest.raises(ValueError, match="value must be an"):
        hp_module.Fixed("fixed", None)


def test_fixed_repr():
    assert repr(hp_module.Fixed("fixed", "value")) == repr(
        hp_module.Fixed("fixed", "value")
    )


def test_fixed_values_property():
    assert list(hp_module.Fixed("fixed", 2).values) == [2]



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np
import pytest
import tensorflow as tf

import keras_tuner
from keras_tuner.engine import tuner_utils
from keras_tuner.tuners import hyperband as hyperband_module


def build_model(hp):
    model = tf.keras.Sequential()
    for i in range(hp.Int("layers", 1, 3)):
        model.add(
            tf.keras.layers.Dense(
                hp.Int(f"units{str(i)}", 1, 5), activation="relu"
            )
        )

        model.add(
            tf.keras.layers.Lambda(
                lambda x: x + hp.Float(f"bias{str(i)}", -1, 1)
            )
        )

    model.add(tf.keras.layers.Dense(1, activation="sigmoid"))
    model.compile("sgd", "mse")
    return model


def test_hyperband_oracle_bracket_configs(tmp_path):
    oracle = hyperband_module.HyperbandOracle(
        objective=keras_tuner.Objective("score", "max"),
        hyperband_iterations=1,
        max_epochs=8,
        factor=2,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    # 8, 4, 2, 1 starting epochs.
    assert oracle._get_num_brackets() == 4

    assert oracle._get_num_rounds(bracket_num=3) == 4
    assert oracle._get_size(bracket_num=3, round_num=0) == 8
    assert oracle._get_epochs(bracket_num=3, round_num=0) == 1
    assert oracle._get_size(bracket_num=3, round_num=3) == 1
    assert oracle._get_epochs(bracket_num=3, round_num=3) == 8

    assert oracle._get_num_rounds(bracket_num=0) == 1
    assert oracle._get_size(bracket_num=0, round_num=0) == 4
    assert oracle._get_epochs(bracket_num=0, round_num=0) == 8


def test_hyperband_oracle_one_sweep_single_thread(tmp_path):
    hp = keras_tuner.HyperParameters()
    hp.Float("a", -100, 100)
    hp.Float("b", -100, 100)
    oracle = hyperband_module.HyperbandOracle(
        hyperparameters=hp,
        objective=keras_tuner.Objective("score", "max"),
        hyperband_iterations=1,
        max_epochs=9,
        factor=3,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    score = 0
    for bracket_num in reversed(range(oracle._get_num_brackets())):
        for round_num in range(oracle._get_num_rounds(bracket_num)):
            for _ in range(oracle._get_size(bracket_num, round_num)):
                trial = oracle.create_trial("tuner0")
                assert trial.status == "RUNNING"
                score += 1
                oracle.update_trial(trial.trial_id, {"score": score})
                trial.status = "COMPLETED"
                oracle.end_trial(trial)
            assert len(
                oracle._brackets[0]["rounds"][round_num]
            ) == oracle._get_size(bracket_num, round_num)
        assert len(oracle._brackets) == 1

    # Iteration should now be complete.
    trial = oracle.create_trial("tuner0")
    assert trial.status == "STOPPED", oracle.hyperband_iterations
    assert len(oracle.ongoing_trials) == 0

    # Brackets should all be finished and removed.
    assert len(oracle._brackets) == 0

    best_trial = oracle.get_best_trials()[0]
    assert best_trial.score == score


def test_hyperband_oracle_one_sweep_parallel(tmp_path):
    hp = keras_tuner.HyperParameters()
    hp.Float("a", -100, 100)
    hp.Float("b", -100, 100)
    oracle = hyperband_module.HyperbandOracle(
        hyperparameters=hp,
        objective=keras_tuner.Objective("score", "max"),
        hyperband_iterations=1,
        max_epochs=4,
        factor=2,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    # All round 0 trials from different brackets can be run
    # in parallel.
    round0_trials = []
    for i in range(10):
        t = oracle.create_trial(f"tuner{str(i)}")
        assert t.status == "RUNNING"
        round0_trials.append(t)

    assert len(oracle._brackets) == 3

    # Round 1 can't be run until enough models from round 0
    # have completed.
    t = oracle.create_trial("tuner10")
    assert t.status == "IDLE"

    for t in round0_trials:
        oracle.update_trial(t.trial_id, {"score": 1})
        t.status = "COMPLETED"
        oracle.end_trial(t)

    round1_trials = []
    for i in range(4):
        t = oracle.create_trial(f"tuner{str(i)}")
        assert t.status == "RUNNING"
        round1_trials.append(t)

    # Bracket 0 is complete as it only has round 0.
    assert len(oracle._brackets) == 2

    # Round 2 can't be run until enough models from round 1
    # have completed.
    t = oracle.create_trial("tuner10")
    assert t.status == "IDLE"

    for t in round1_trials:
        oracle.update_trial(t.trial_id, {"score": 1})
        t.status = "COMPLETED"
        oracle.end_trial(t)

    # Only one trial runs in round 2.
    round2_trial = oracle.create_trial("tuner0")

    assert len(oracle._brackets) == 1

    # No more trials to run, but wait for existing brackets to end.
    t = oracle.create_trial("tuner10")
    assert t.status == "IDLE"

    oracle.update_trial(round2_trial.trial_id, {"score": 1})
    round2_trial.status = "COMPLETED"
    oracle.end_trial(round2_trial)

    t = oracle.create_trial("tuner10")
    assert t.status == "STOPPED", oracle._current_sweep


def test_hyperband_integration(tmp_path):
    tuner = hyperband_module.Hyperband(
        objective="val_loss",
        hypermodel=build_model,
        hyperband_iterations=2,
        max_epochs=6,
        factor=3,
        directory=tmp_path,
    )

    x, y = np.ones((2, 5)), np.ones((2, 1))
    tuner.search(x, y, validation_data=(x, y))

    # Make sure Oracle is registering new HPs.
    updated_hps = tuner.oracle.get_space().values
    assert "units1" in updated_hps
    assert "bias1" in updated_hps

    tf.get_logger().setLevel(logging.ERROR)

    best_score = tuner.oracle.get_best_trials()[0].score
    best_model = tuner.get_best_models()[0]
    assert best_model.evaluate(x, y) == best_score


def test_hyperband_save_and_restore(tmp_path):
    tuner = hyperband_module.Hyperband(
        objective="val_loss",
        hypermodel=build_model,
        hyperband_iterations=1,
        max_epochs=7,
        factor=2,
        directory=tmp_path,
    )

    x, y = np.ones((2, 5)), np.ones((2, 1))
    tuner.search(x, y, validation_data=(x, y))

    num_trials = len(tuner.oracle.trials)
    assert num_trials > 0
    assert tuner.oracle._current_iteration == 0

    tuner.save()
    tuner.trials = {}
    tuner.oracle._current_iteration = 0
    tuner.reload()

    assert len(tuner.oracle.trials) == num_trials
    assert tuner.oracle._current_iteration == 0


def test_hyperband_load_weights(tmp_path):
    tuner = hyperband_module.Hyperband(
        objective="val_loss",
        hypermodel=build_model,
        hyperband_iterations=1,
        max_epochs=2,
        factor=2,
        directory=tmp_path,
    )
    x, y = np.ones((2, 5)), np.ones((2, 1))
    nb_brackets = tuner.oracle._get_num_brackets()
    assert nb_brackets == 2
    nb_models_round_0 = tuner.oracle._get_size(bracket_num=1, round_num=0)
    assert nb_models_round_0 == 2
    # run the trials for the round 0 (from scratch)
    for _ in range(nb_models_round_0):
        trial = tuner.oracle.create_trial("tuner0")
        result = tuner.run_trial(trial, x, y, validation_data=(x, y))
        tuner.oracle.update_trial(
            trial.trial_id,
            tuner_utils.convert_to_metrics_dict(result, tuner.oracle.objective),
            tuner_utils.get_best_step(result, tuner.oracle.objective),
        )
        trial.status = "COMPLETED"
        tuner.oracle.end_trial(trial)

    # ensure the model run in round 1 is loaded from the best model in round 0
    trial = tuner.oracle.create_trial("tuner0")
    hp = trial.hyperparameters
    assert "tuner/trial_id" in hp
    new_model = tuner._try_build(hp)
    assert new_model.predict(x).shape == y.shape
    # get new model weights
    new_model_weights = new_model.weights.copy()
    # get weights from the best model in round 0
    best_trial_round_0_id = hp["tuner/trial_id"]
    best_hp_round_0 = tuner.oracle.trials[best_trial_round_0_id].hyperparameters
    best_model_round_0 = tuner._try_build(best_hp_round_0)
    best_model_round_0.load_weights(
        tuner._get_checkpoint_fname(best_trial_round_0_id)
    )
    assert best_model_round_0.predict(x).shape == y.shape
    best_model_round_0_weights = best_model_round_0.weights.copy()
    # compare the weights
    assert len(new_model_weights) == len(best_model_round_0_weights)
    assert all(
        np.alltrue(new_weight == best_old_weight)
        for new_weight, best_old_weight in zip(
            new_model_weights, best_model_round_0_weights
        )
    )


def test_factor_less_than_2_error():
    with pytest.raises(ValueError, match="factor needs"):
        hyperband_module.HyperbandOracle(
            objective=keras_tuner.Objective("score", "max"),
            hyperband_iterations=1,
            max_epochs=8,
            factor=1,
        )


def test_exhausted_values_during_init_bracket(tmp_path):
    oracle = hyperband_module.HyperbandOracle(
        objective=keras_tuner.Objective("score", "max"),
        hyperband_iterations=1,
        max_epochs=8,
        factor=3,
    )
    oracle._set_project_dir(tmp_path, "untitled")
    hp = oracle.get_space()
    hp.Boolean("bool")
    oracle.update_space(hp)

    trial_1 = oracle.create_trial("a")
    trial_2 = oracle.create_trial("b")
    trial_3 = oracle.create_trial("c")
    assert trial_3.status == "IDLE"
    print(oracle.trials)

    oracle.update_trial(
        trial_id=trial_1.trial_id,
        metrics={oracle.objective.name: np.random.rand()},
    )
    oracle.end_trial(trial_1)

    oracle.update_trial(
        trial_id=trial_2.trial_id,
        metrics={oracle.objective.name: np.random.rand()},
    )
    oracle.end_trial(trial_2)

    trial_1 = oracle.create_trial("a")
    assert trial_1.status == "STOPPED"



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import pytest
from sklearn import datasets
from sklearn import decomposition
from sklearn import ensemble
from sklearn import linear_model
from sklearn import metrics
from sklearn import model_selection
from sklearn import neighbors
from sklearn import pipeline

import keras_tuner


def build_model(hp):
    model_type = hp.Choice("model_type", ["random_forest", "ridge", "knn"])
    if model_type == "random_forest":
        with hp.conditional_scope("model_type", "random_forest"):
            model = ensemble.RandomForestClassifier(
                n_estimators=hp.Int("n_estimators", 10, 50, step=10),
                max_depth=hp.Int("max_depth", 3, 10),
            )
    elif model_type == "ridge":
        with hp.conditional_scope("model_type", "ridge"):
            model = linear_model.RidgeClassifier(
                alpha=hp.Float("alpha", 1e-3, 1, sampling="log")
            )
    elif model_type == "knn":
        with hp.conditional_scope("model_type", "knn"):
            k = hp.Int("n_neighbors", 1, 30, default=5)
            model = neighbors.KNeighborsClassifier(
                n_neighbors=k,
                weights=hp.Choice(
                    "weights", ["uniform", "distance"], default="uniform"
                ),
            )
    else:
        raise ValueError("Unrecognized model_type")
    return model


def build_pipeline(hp):
    n_components = hp.Choice("n_components", [2, 5, 10], default=5)
    pca = decomposition.PCA(n_components=n_components)

    model_type = hp.Choice("model_type", ["random_forest", "ridge", "knn"])
    if model_type == "random_forest":
        with hp.conditional_scope("model_type", "random_forest"):
            model = ensemble.RandomForestClassifier(
                n_estimators=hp.Int("n_estimators", 10, 50, step=10),
                max_depth=hp.Int("max_depth", 3, 10),
            )
    elif model_type == "ridge":
        with hp.conditional_scope("model_type", "ridge"):
            model = linear_model.RidgeClassifier(
                alpha=hp.Float("alpha", 1e-3, 1, sampling="log")
            )
    elif model_type == "knn":
        with hp.conditional_scope("model_type", "knn"):
            k = hp.Int("n_neighbors", 1, 30, default=5)
            model = neighbors.KNeighborsClassifier(
                n_neighbors=k,
                weights=hp.Choice(
                    "weights", ["uniform", "distance"], default="uniform"
                ),
            )
    else:
        raise ValueError("Unrecognized model_type")

    skpipeline = pipeline.Pipeline([("pca", pca), ("clf", model)])
    return skpipeline


def test_sklearn_tuner_simple_with_np(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        directory=tmp_path,
    )

    x = np.random.uniform(size=(50, 10))
    y = np.random.randint(0, 2, size=(50,))
    tuner.search(x, y)

    assert len(tuner.oracle.trials) == 10

    best_trial = tuner.oracle.get_best_trials()[0]
    assert best_trial.status == "COMPLETED"
    assert best_trial.score is not None
    assert best_trial.best_step == 0
    assert best_trial.metrics.exists("score")

    # Make sure best model can be reloaded.
    best_model = tuner.get_best_models()[0]
    best_model.score(x, y)


@pytest.mark.filterwarnings("ignore:.*column-vector")
def test_sklearn_tuner_with_df(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        directory=tmp_path,
    )

    x = pd.DataFrame(np.random.uniform(size=(50, 10)))
    y = pd.DataFrame(np.random.randint(0, 2, size=(50,)))
    tuner.search(x, y)

    assert len(tuner.oracle.trials) == 10


def test_sklearn_custom_scoring_and_cv(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        scoring=metrics.make_scorer(metrics.balanced_accuracy_score),
        cv=model_selection.StratifiedKFold(5),
        directory=tmp_path,
    )

    x = np.random.uniform(size=(50, 10))
    y = np.random.randint(0, 2, size=(50,))
    tuner.search(x, y)

    assert len(tuner.oracle.trials) == 10

    best_trial = tuner.oracle.get_best_trials()[0]
    assert best_trial.status == "COMPLETED"
    assert best_trial.score is not None
    assert best_trial.best_step == 0
    assert best_trial.metrics.exists("score")

    # Make sure best model can be reloaded.
    best_model = tuner.get_best_models()[0]
    best_model.score(x, y)


def test_sklearn_additional_metrics(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        metrics=[metrics.balanced_accuracy_score, metrics.recall_score],
        directory=tmp_path,
    )

    x = np.random.uniform(size=(50, 10))
    y = np.random.randint(0, 2, size=(50,))
    tuner.search(x, y)

    assert len(tuner.oracle.trials) == 10

    best_trial = tuner.oracle.get_best_trials()[0]
    assert best_trial.status == "COMPLETED"
    assert best_trial.score is not None
    assert best_trial.best_step == 0
    assert best_trial.metrics.exists("score")
    assert best_trial.metrics.exists("balanced_accuracy_score")
    assert best_trial.metrics.exists("recall_score")

    # Make sure best model can be reloaded.
    best_model = tuner.get_best_models()[0]
    best_model.score(x, y)


def test_sklearn_sample_weight(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        directory=tmp_path,
    )

    x = np.random.uniform(size=(50, 10))
    y = np.random.randint(0, 2, size=(50,))
    sample_weight = np.random.uniform(0.1, 1, size=(50,))
    tuner.search(x, y, sample_weight=sample_weight)

    assert len(tuner.oracle.trials) == 10

    best_trial = tuner.oracle.get_best_trials()[0]
    assert best_trial.status == "COMPLETED"
    assert best_trial.score is not None
    assert best_trial.best_step == 0
    assert best_trial.metrics.exists("score")

    # Make sure best model can be reloaded.
    best_model = tuner.get_best_models()[0]
    best_model.score(x, y)


def test_sklearn_pipeline(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_pipeline,
        directory=tmp_path,
    )

    x = np.random.uniform(size=(50, 10))
    y = np.random.randint(0, 2, size=(50,))
    sample_weight = np.random.uniform(0.1, 1, size=(50,))
    tuner.search(x, y, sample_weight=sample_weight)

    assert len(tuner.oracle.trials) == 10

    best_trial = tuner.oracle.get_best_trials()[0]
    assert best_trial.status == "COMPLETED"
    assert best_trial.score is not None
    assert best_trial.best_step == 0
    assert best_trial.metrics.exists("score")

    # Make sure best pipeline can be reloaded.
    best_pipeline = tuner.get_best_models()[0]
    best_pipeline.score(x, y)


def test_sklearn_cv_with_groups(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        cv=model_selection.GroupKFold(5),
        directory=tmp_path,
    )

    x = np.random.uniform(size=(50, 10))
    y = np.random.randint(0, 2, size=(50,))
    groups = np.random.randint(0, 5, size=(50,))
    tuner.search(x, y, groups=groups)

    assert len(tuner.oracle.trials) == 10

    best_trial = tuner.oracle.get_best_trials()[0]
    assert best_trial.status == "COMPLETED"
    assert best_trial.score is not None
    assert best_trial.best_step == 0
    assert best_trial.metrics.exists("score")

    # Make sure best model can be reloaded.
    best_model = tuner.get_best_models()[0]
    best_model.score(x, y)


def test_sklearn_real_data(tmp_path):
    tuner = keras_tuner.SklearnTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        ),
        hypermodel=build_model,
        scoring=metrics.make_scorer(metrics.accuracy_score),
        metrics=metrics.accuracy_score,
        cv=model_selection.StratifiedKFold(5),
        directory=tmp_path,
    )

    x, y = datasets.load_iris(return_X_y=True)
    x_train, x_test, y_train, y_test = model_selection.train_test_split(
        x, y, test_size=0.2
    )

    tuner.search(x_train, y_train)

    best_models = tuner.get_best_models(10)
    best_model = best_models[0]
    worst_model = best_models[9]
    best_model_score = best_model.score(x_test, y_test)
    worst_model_score = worst_model.score(x_test, y_test)

    assert best_model_score > 0.8
    assert best_model_score >= worst_model_score


def test_sklearn_not_install_error(tmp_path):
    sklearn_module = keras_tuner.tuners.sklearn_tuner.sklearn
    keras_tuner.tuners.sklearn_tuner.sklearn = None

    with pytest.raises(ImportError, match="Please install sklearn"):
        keras_tuner.SklearnTuner(
            oracle=keras_tuner.oracles.BayesianOptimizationOracle(
                objective=keras_tuner.Objective("score", "max"), max_trials=10
            ),
            hypermodel=build_model,
            directory=tmp_path,
        )

    keras_tuner.tuners.sklearn_tuner.sklearn = sklearn_module


def test_sklearn_wrong_data_type(tmp_path):
    with pytest.raises(RuntimeError, match="Expected the data to be numpy"):
        tuner = keras_tuner.SklearnTuner(
            oracle=keras_tuner.oracles.BayesianOptimizationOracle(
                objective=keras_tuner.Objective("score", "max"), max_trials=10
            ),
            hypermodel=build_model,
            scoring=metrics.make_scorer(metrics.accuracy_score),
            cv=model_selection.StratifiedKFold(3),
            directory=tmp_path,
        )

        tuner.search([1, 2, 3, 4, 5, 6], [1, 1, 1, 2, 2, 2])



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for HyperResNet Model."""

import os

import numpy as np
import pytest
import tensorflow as tf

from keras_tuner.applications import resnet
from keras_tuner.engine import hyperparameters as hp_module


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Causes CI to stall")
@pytest.mark.parametrize("version", ["v1", "v2", "next"])
def test_model_construction(version):
    hp = hp_module.HyperParameters()
    hp.Choice("version", [version])
    hypermodel = resnet.HyperResNet(input_shape=(128, 128, 3), classes=10)
    model = hypermodel.build(hp)
    assert hp.values["version"] == version
    assert model.layers
    assert model.name == "ResNet"
    assert model.output_shape == (None, 10)
    model.train_on_batch(np.ones((1, 128, 128, 3)), np.ones((1, 10)))
    out = model.predict(np.ones((1, 128, 128, 3)))
    assert out.shape == (1, 10)


def test_hyperparameter_existence_and_defaults():
    hp = hp_module.HyperParameters()
    hypermodel = resnet.HyperResNet(input_shape=(256, 256, 3), classes=10)
    hypermodel.build(hp)
    assert hp.get("version") == "v2"
    assert hp.get("conv3_depth") == 4
    assert hp.get("conv4_depth") == 6
    assert hp.get("learning_rate") == 0.01
    assert hp.get("pooling") == "avg"


def test_include_top_false():
    hp = hp_module.HyperParameters()
    hypermodel = resnet.HyperResNet(
        input_shape=(256, 256, 3), classes=10, include_top=False
    )
    model = hypermodel.build(hp)
    # Check that model wasn't compiled.
    assert not model.optimizer


def test_hyperparameter_override():
    hp = hp_module.HyperParameters()
    hp.Choice("version", ["v1"])
    hp.Fixed("conv3_depth", 10)
    hypermodel = resnet.HyperResNet(input_shape=(256, 256, 3), classes=10)
    hypermodel.build(hp)
    assert hp.get("version") == "v1"
    assert hp.get("conv3_depth") == 10
    assert hp.get("conv4_depth") == 6


def test_input_tensor():
    hp = hp_module.HyperParameters()
    inputs = tf.keras.Input(shape=(256, 256, 3))
    hypermodel = resnet.HyperResNet(input_tensor=inputs, include_top=False)
    model = hypermodel.build(hp)
    assert model.inputs == [inputs]


def test_pooling_is_max():
    hp = hp_module.HyperParameters()
    hp.values["pooling"] = "max"
    hypermodel = resnet.HyperResNet(input_shape=(32, 32, 3), classes=10)
    hypermodel.build(hp)


def test_no_classes_raise_error():
    with pytest.raises(ValueError, match="classes"):
        resnet.HyperResNet(input_shape=(32, 32, 3))


def test_no_input_shape_tensor_raise_error():
    with pytest.raises(ValueError, match="input_tensor"):
        resnet.HyperResNet(classes=10)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Tuner base class."


import copy
import os
import traceback
import warnings

import tensorflow as tf

from keras_tuner import config as config_module
from keras_tuner import errors
from keras_tuner import utils
from keras_tuner.distribute import utils as dist_utils
from keras_tuner.engine import hypermodel as hm_module
from keras_tuner.engine import oracle as oracle_module
from keras_tuner.engine import stateful
from keras_tuner.engine import trial as trial_module
from keras_tuner.engine import tuner_utils


class BaseTuner(stateful.Stateful):
    """Tuner base class.

    `BaseTuner` is the super class of all `Tuner` classes. It defines the APIs
    for the `Tuner` classes and serves as a wrapper class for the internal
    logics.

    `BaseTuner` supports parallel tuning. In parallel tuning, the communication
    between `BaseTuner` and `Oracle` are all going through gRPC. There are
    multiple running instances of `BaseTuner` but only one `Oracle`. This design
    allows the user to run the same script on multiple machines to launch the
    parallel tuning.

    The `Oracle` instance should manage the life cycles of all the `Trial`s,
    while a `BaseTuner` is a worker for running the `Trial`s. `BaseTuner`s
    requests `Trial`s from the `Oracle`, run them, and report the results back
    to the `Oracle`. A `BaseTuner` also handles events happening during running
    the `Trial`, like saving the model, logging, error handling. Other than
    these responsibilities, a `BaseTuner` should avoid managing a `Trial` since
    the relevant contexts for a `Trial` are in the `Oracle`, which only
    accessible from gRPC.

    The `BaseTuner` should be a general tuner for all types of models and avoid
    any logic directly related to Keras. The Keras related logics should be
    handled by the `Tuner` class, which is a subclass of `BaseTuner`.

    Args:
        oracle: Instance of Oracle class.
        hypermodel: Instance of `HyperModel` class (or callable that takes
            hyperparameters and returns a `Model` instance). It is optional
            when `Tuner.run_trial()` is overriden and does not use
            `self.hypermodel`.
        directory: A string, the relative path to the working directory.
        project_name: A string, the name to use as prefix for files saved by
            this Tuner.
        overwrite: Boolean, defaults to `False`. If `False`, reloads an
            existing project of the same name if one is found. Otherwise,
            overwrites the project.
        **kwargs: Arguments for backward compatibility.

    Attributes:
        remaining_trials: Number of trials remaining, `None` if `max_trials` is
            not set. This is useful when resuming a previously stopped search.
    """

    def __init__(
        self,
        oracle,
        hypermodel=None,
        directory=None,
        project_name=None,
        overwrite=False,
        **kwargs,
    ):
        if not isinstance(oracle, oracle_module.Oracle):
            raise ValueError(
                "Expected `oracle` argument to be an instance of `Oracle`. "
                f"Received: oracle={oracle} (of type ({type(oracle)})."
            )

        logger = kwargs.pop("logger", None)
        if logger is not None:
            warnings.warn(
                "The `logger` argument in `BaseTuner.__init__() is "
                "no longer supported and will be ignored.",
                DeprecationWarning,
                stacklevel=2,
            )
        self.logger = logger

        if len(kwargs) > 0:
            raise ValueError(
                f"Unrecognized arguments {list(kwargs.keys())} "
                "for `BaseTuner.__init__()`."
            )

        self.oracle = oracle
        self.hypermodel = hm_module.get_hypermodel(hypermodel)

        # Ops and metadata
        self.directory = directory or "."
        self.project_name = project_name or "untitled_project"
        self.oracle._set_project_dir(self.directory, self.project_name)

        if overwrite and tf.io.gfile.exists(self.project_dir):
            tf.io.gfile.rmtree(self.project_dir)

        # To support tuning distribution.
        self.tuner_id = os.environ.get("KERASTUNER_TUNER_ID", "tuner0")

        # Reloading state.
        if not overwrite and tf.io.gfile.exists(self._get_tuner_fname()):
            tf.get_logger().info(
                f"Reloading Tuner from {self._get_tuner_fname()}"
            )
            self.reload()
        else:
            # Only populate initial space if not reloading.
            self._populate_initial_space()

        # Run in distributed mode.
        if dist_utils.is_chief_oracle():
            # Blocks forever.
            # Avoid import at the top, to avoid inconsistent protobuf versions.
            from keras_tuner.distribute import oracle_chief

            oracle_chief.start_server(self.oracle)
        elif dist_utils.has_chief_oracle():
            # Proxies requests to the chief oracle.
            # Avoid import at the top, to avoid inconsistent protobuf versions.
            from keras_tuner.distribute import oracle_client

            self.oracle = oracle_client.OracleClient(self.oracle)

        # In parallel tuning, everything below in __init__() is for workers
        # only.
        # Logs etc
        self._display = tuner_utils.Display(oracle=self.oracle)

    def _activate_all_conditions(self):
        # Lists of stacks of conditions used during `explore_space()`.
        scopes_never_active = []
        scopes_once_active = []

        hp = self.oracle.get_space()
        while True:
            self.hypermodel.build(hp)
            self.oracle.update_space(hp)

            # Update the recorded scopes.
            for conditions in hp.active_scopes:
                if conditions not in scopes_once_active:
                    scopes_once_active.append(copy.deepcopy(conditions))
                if conditions in scopes_never_active:
                    scopes_never_active.remove(conditions)
            for conditions in hp.inactive_scopes:
                if conditions not in scopes_once_active:
                    scopes_never_active.append(copy.deepcopy(conditions))

            # All conditional scopes are activated.
            if not scopes_never_active:
                break

            # Generate new values to activate new conditions.
            hp = self.oracle.get_space()
            conditions = scopes_never_active[0]
            for condition in conditions:
                hp.values[condition.name] = condition.values[0]

            hp.ensure_active_values()

    def _populate_initial_space(self):
        """Populate initial search space for oracle.

        Keep this function as a subroutine for AutoKeras to override. The space
        may not be ready at the initialization of the tuner, but after seeing
        the training data.

        Build hypermodel multiple times to find all conditional hps. It
        generates hp values based on the not activated `conditional_scopes`
        found in the builds.
        """
        if self.hypermodel is None:
            return

        # declare_hyperparameters is not overriden.
        hp = self.oracle.get_space()
        self.hypermodel.declare_hyperparameters(hp)
        self.oracle.update_space(hp)
        self._activate_all_conditions()

    def search(self, *fit_args, **fit_kwargs):
        """Performs a search for best hyperparameter configuations.

        Args:
            *fit_args: Positional arguments that should be passed to
              `run_trial`, for example the training and validation data.
            **fit_kwargs: Keyword arguments that should be passed to
              `run_trial`, for example the training and validation data.
        """
        if "verbose" in fit_kwargs:
            self._display.verbose = fit_kwargs.get("verbose")
        self.on_search_begin()
        while True:
            self.pre_create_trial()
            trial = self.oracle.create_trial(self.tuner_id)
            if trial.status == trial_module.TrialStatus.STOPPED:
                # Oracle triggered exit.
                tf.get_logger().info("Oracle triggered exit")
                break
            if trial.status == trial_module.TrialStatus.IDLE:
                # Oracle is calculating, resend request.
                continue

            self.on_trial_begin(trial)
            self._try_run_and_update_trial(trial, *fit_args, **fit_kwargs)
            self.on_trial_end(trial)
        self.on_search_end()

    def _run_and_update_trial(self, trial, *fit_args, **fit_kwargs):
        results = self.run_trial(trial, *fit_args, **fit_kwargs)
        if self.oracle.get_trial(trial.trial_id).metrics.exists(
            self.oracle.objective.name
        ):
            # The oracle is updated by calling `self.oracle.update_trial()` in
            # `Tuner.run_trial()`. For backward compatibility, we support this
            # use case. No further action needed in this case.
            warnings.warn(
                "The use case of calling "
                "`self.oracle.update_trial(trial_id, metrics)` "
                "in `Tuner.run_trial()` to report the metrics is deprecated, "
                "and will be removed in the future."
                "Please remove the call and do 'return metrics' "
                "in `Tuner.run_trial()` instead. ",
                DeprecationWarning,
                stacklevel=2,
            )
            return

        tuner_utils.validate_trial_results(
            results, self.oracle.objective, "Tuner.run_trial()"
        ),
        self.oracle.update_trial(
            trial.trial_id,
            # Convert to dictionary before calling `update_trial()`
            # to pass it from gRPC.
            tuner_utils.convert_to_metrics_dict(
                results,
                self.oracle.objective,
            ),
            step=tuner_utils.get_best_step(results, self.oracle.objective),
        )

    def _try_run_and_update_trial(self, trial, *fit_args, **fit_kwargs):
        try:
            self._run_and_update_trial(trial, *fit_args, **fit_kwargs)
            trial.status = trial_module.TrialStatus.COMPLETED
            return
        except Exception as e:
            if isinstance(e, errors.FatalError):
                raise e
            if config_module.DEBUG:
                # Printing the stacktrace and the error.
                traceback.print_exc()

            if isinstance(e, errors.FailedTrialError):
                trial.status = trial_module.TrialStatus.FAILED
            else:
                trial.status = trial_module.TrialStatus.INVALID

            # Include the stack traces in the message.
            message = traceback.format_exc()
            trial.message = message

    def run_trial(self, trial, *fit_args, **fit_kwargs):
        """Evaluates a set of hyperparameter values."""
        raise NotImplementedError

    def save_model(self, trial_id, model, step=0):
        """Saves a Model for a given trial.

        Args:
            trial_id: The ID of the `Trial` corresponding to this Model.
            model: The trained model.
            step: Integer, for models that report intermediate results to the
                `Oracle`, the step the saved file correspond to. For example,
                for Keras models this is the number of epochs trained.
        """
        raise NotImplementedError

    def load_model(self, trial):
        """Loads a Model from a given trial.

        For models that report intermediate results to the `Oracle`, generally
        `load_model` should load the best reported `step` by relying of
        `trial.best_step`.

        Args:
            trial: A `Trial` instance, the `Trial` corresponding to the model
                to load.
        """
        raise NotImplementedError

    def pre_create_trial(self):
        """Called before self.oracle.create_trial and before on_trial_begin."""

    def on_trial_begin(self, trial):
        """Called at the beginning of a trial.

        Args:
            trial: A `Trial` instance.
        """
        self._display.on_trial_begin(self.oracle.get_trial(trial.trial_id))

    def on_trial_end(self, trial):
        """Called at the end of a trial.

        Args:
            trial: A `Trial` instance.
        """
        self.oracle.end_trial(trial)
        # Display needs the updated trial scored by the Oracle.
        self._display.on_trial_end(self.oracle.get_trial(trial.trial_id))
        self.save()

    def on_search_begin(self):
        """Called at the beginning of the `search` method."""
        pass

    def on_search_end(self):
        """Called at the end of the `search` method."""
        pass

    def get_best_models(self, num_models=1):
        """Returns the best model(s), as determined by the objective.

        This method is for querying the models trained during the search.
        For best performance, it is recommended to retrain your Model on the
        full dataset using the best hyperparameters found during `search`,
        which can be obtained using `tuner.get_best_hyperparameters()`.

        Args:
            num_models: Optional number of best models to return.
                Defaults to 1.

        Returns:
            List of trained models sorted from the best to the worst.
        """
        best_trials = self.oracle.get_best_trials(num_models)
        models = [self.load_model(trial) for trial in best_trials]
        return models

    def get_best_hyperparameters(self, num_trials=1):
        """Returns the best hyperparameters, as determined by the objective.

        This method can be used to reinstantiate the (untrained) best model
        found during the search process.

        Example:

        ```python
        best_hp = tuner.get_best_hyperparameters()[0]
        model = tuner.hypermodel.build(best_hp)
        ```

        Args:
            num_trials: Optional number of `HyperParameters` objects to return.

        Returns:
            List of `HyperParameter` objects sorted from the best to the worst.
        """
        return [
            t.hyperparameters for t in self.oracle.get_best_trials(num_trials)
        ]

    def search_space_summary(self, extended=False):
        """Print search space summary.

        The methods prints a summary of the hyperparameters in the search
        space, which can be called before calling the `search` method.

        Args:
            extended: Optional boolean, whether to display an extended summary.
                Defaults to False.
        """
        print("Search space summary")
        hp = self.oracle.get_space()
        print(f"Default search space size: {len(hp.space)}")
        for p in hp.space:
            config = p.get_config()
            name = config.pop("name")
            print(f"{name} ({p.__class__.__name__})")
            print(config)

    def results_summary(self, num_trials=10):
        """Display tuning results summary.

        The method prints a summary of the search results including the
        hyperparameter values and evaluation results for each trial.

        Args:
            num_trials: Optional number of trials to display. Defaults to 10.
        """
        print("Results summary")
        print(f"Results in {self.project_dir}")
        print("Showing %d best trials" % num_trials)
        print(f"{self.oracle.objective}")

        best_trials = self.oracle.get_best_trials(num_trials)
        for trial in best_trials:
            print()
            trial.summary()

    @property
    def remaining_trials(self):
        """Returns the number of trials remaining.

        Will return `None` if `max_trials` is not set. This is useful when
        resuming a previously stopped search.
        """
        return self.oracle.remaining_trials()

    def get_state(self):
        return {}

    def set_state(self, state):
        pass

    def _is_worker(self):
        """Return true only if in parallel tuning and is a worker tuner."""
        return (
            dist_utils.has_chief_oracle() and not dist_utils.is_chief_oracle()
        )

    def save(self):
        """Saves this object to its project directory."""
        if not self._is_worker():
            self.oracle.save()
        super().save(self._get_tuner_fname())

    def reload(self):
        """Reloads this object from its project directory."""
        if not self._is_worker():
            self.oracle.reload()
        super().reload(self._get_tuner_fname())

    @property
    def project_dir(self):
        dirname = os.path.join(str(self.directory), self.project_name)
        utils.create_directory(dirname)
        return dirname

    def get_trial_dir(self, trial_id):
        dirname = os.path.join(str(self.project_dir), f"trial_{str(trial_id)}")
        utils.create_directory(dirname)
        return dirname

    def _get_tuner_fname(self):
        return os.path.join(str(self.project_dir), f"{str(self.tuner_id)}.json")



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

import pytest

from keras_tuner.distribute import utils
from keras_tuner.test_utils import mock_distribute


def test_no_port_error():
    with mock.patch.object(os, "environ", mock_distribute.MockEnvVars()):
        os.environ["KERASTUNER_ORACLE_IP"] = "127.0.0.1"
        os.environ["KERASTUNER_TUNER_ID"] = "worker0"
        with pytest.raises(RuntimeError, match="KERASTUNER_ORACLE_PORT"):
            utils.has_chief_oracle()


def test_no_id_error():
    with mock.patch.object(os, "environ", mock_distribute.MockEnvVars()):
        os.environ["KERASTUNER_ORACLE_IP"] = "127.0.0.1"
        os.environ["KERASTUNER_ORACLE_PORT"] = "80"
        with pytest.raises(RuntimeError, match="KERASTUNER_TUNER_ID"):
            utils.has_chief_oracle()
